/*
	AEFlame - After Effects version of flame
	Original version of flame by Scott Draves <spot@cs.cmu.edu>
	After Effects version by Andrew B. Davidson <aeflame@mindspring.com> 
    flame Copyright (C) 1992 Scott Draves.  
    AEFlame Copyright (C) 1999 Andrew B. Davidson.  

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "AEFlame.h"
#include "Param_Utils.h"
#include "AE_Macros.h"
#include <stdio.h>		// various string functions (strcpy, sprintf, etc)
#include <string.h>		// various memory functions (memset, memmove, etc.)
#include "flame.h"
#include "cmap.h"
#include <time.h>		// for time() function to dynamically re-seed random numbers
//#include "PF_Suite_Helper.h"

// PROFILING
//#include <profiler.h>

// platform dependent declarations
int GetNthFile(const char *iDirName, const char *iMatchPattern, long iFileIndex, char *oPath);



static int gNumCmaps = 0;
static int gDisplayedAboutBox = TRUE;		// set this to FALSE to make it display about box on first render

#define SETTINGS_FILENAME	"C:\\AEFlameSettings.txt"
#define CMAP_DIRECTORY		"C:\\CMAPS\\"

//=========================================================================
// standard after effects plugin main routine
//-------------------------------------------------------------------------
PF_Err main (	PF_Cmd			cmd,
				PF_InData		*in_data,
				PF_OutData		*out_data,
				PF_ParamDef		*params[],
				PF_LayerDef		*output,
				void			*extra	)
{
	PF_Err		err = PF_Err_NONE;
	
	switch (cmd) 
	{
		case PF_Cmd_ABOUT:					// get about box info
			err = About(in_data, out_data, params, output);
			break;
		case PF_Cmd_GLOBAL_SETUP:			// set up global structures
			err = GlobalSetup(in_data, out_data, params, output);
			break;
		case PF_Cmd_PARAMS_SETUP:			// set up our params
			err = ParamsSetup(in_data, out_data, params, output);
			break;
		case PF_Cmd_GLOBAL_SETDOWN:			// dispose of global structures
			err = GlobalSetdown(in_data, out_data, params, output);
			break;
		case PF_Cmd_RENDER:					// render a frame
			err = Render(in_data, out_data, params, output);
			break;
		case PF_Cmd_FRAME_SETUP:			// set up frame structures
			// we don't need to do anything
			break;
		case PF_Cmd_FRAME_SETDOWN:			// dispose of frame structures
			// we don't need to do anything
			break;
		case PF_Cmd_USER_CHANGED_PARAM:		// user has changed a param that had PF_ParamFlag_SUPERVISE flag set
			HandleChangedParam(in_data, out_data, params, output, extra);
			break;	
		case PF_Cmd_DO_DIALOG:
			break;
	}
	return err;
}

//=========================================================================
// about box
//-------------------------------------------------------------------------
PF_Err About (	PF_InData		*in_data,
				PF_OutData		*out_data,
				PF_ParamDef		*params[],
				PF_LayerDef		*output )
{
	PF_SPRINTF(	out_data->return_msg, 
//				"%s %d.%d [%s %s]\r%s",
				"%s [%s %s]\r%s",
				AEF_NAME, 
/*				AEF_MAJOR_VERSION, 
				AEF_MINOR_VERSION, 
*/				__DATE__,
				__TIME__,
				AEF_DESCRIPTION);

	return PF_Err_NONE;
}

static double gExploreVectorVars[kNumXforms][kNumVariations];
static double gExploreVectorCoeffs[kNumXforms][kNumCoefficients];

//=========================================================================
// 
//-------------------------------------------------------------------------
PF_Err GlobalSetup (PF_InData		*in_data,
					PF_OutData		*out_data,
					PF_ParamDef		*params[],
					PF_LayerDef		*output )
{
	PF_Err	err				= PF_Err_NONE;

	out_data->my_version	= PF_VERSION(	AEF_MAJOR_VERSION, 
											AEF_MINOR_VERSION,
											AEF_BUG_VERSION, 
											AEF_STAGE_VERSION, 
											AEF_BUILD_VERSION);

/*	From the SDK documentation:
	PF_OutFlag_USE_OUTPUT_EXTENT
		The output layer is passed with an "extent rect" indicating
		the area of the layer that actually contains visible image data.  If
		the effect changes its behavior based on the extent rect (for instance,
		by not iterating over the entire image), set this flag, so the
		application will know whether having the extent	change should cause
		the frame to re-render.  Specify this flag at PF_Cmd_GLOBAL_SETUP.
*/	out_data->out_flags |= PF_OutFlag_USE_OUTPUT_EXTENT;

/*	From the SDK documentation:
	PF_OutFlag_NON_PARAM_VARY
		Set this if the effect uses information other than the parameters
		in the param list to generate its output at the current time.
		For instance, if the effect uses the current time of the frame
		or some random value to decide the output, set this flag.  This
		flag should be sent at PF_Cmd_GLOBAL_SETUP.  If the effect
		produces changing frames when applied to a still image and
		all parameters are constant, that's a sure sign that this bit
		should be set (e.g. Wave Warp).
*/
	// ABD this should *not* be set because we use the same random sequence every
	// frame (i.e. although we use random numbers, we don't produce changing
	// frames when applied to a still image, because the random numbers we use
	// are the same for every frame... unless the user changes the random seed 
	// parameter each frame)
//	out_data->out_flags |=  PF_OutFlag_NON_PARAM_VARY;
 

/*	From the SDK documentation:
	PF_OutFlag_PIX_INDEPENDENT
		Set this flag if the output at a given pixel is not dependent on the values
		of the pixels around it. If this is set, the pixels After Effects does not
		care about (because of field rendering, for example) could be filled with garbage
		colors.  Please set this flag at PF_Cmd_GLOBAL_SETUP.
	out_data->out_flags |=  PF_OutFlag_PIX_INDEPENDENT;
*/	
	

	{
		int xform, i;
		for ( xform = 0 ; xform < kNumXforms ; xform++ )
		{
			for ( i = 0 ; i < kNumVariations ; i++ )
				gExploreVectorVars[xform][i] = 0.0;
			for ( i = 0 ; i < kNumCoefficients ; i++ )
				gExploreVectorCoeffs[xform][i] = 0.0;
		}
	}

// PROFILING
//ProfilerInit(collectDetailed, bestTimeBase, 20, 5);

	return err;
}



//=========================================================================
// 
//-------------------------------------------------------------------------
PF_Err GlobalSetdown (	PF_InData		*in_data,
						PF_OutData		*out_data,
						PF_ParamDef		*params[],
						PF_LayerDef		*output )
{
// PROFILING
//ProfilerDump("\pAEFlame.prof");
//ProfilerTerm();
	return PF_Err_NONE;
}

//=========================================================================
// 
//-------------------------------------------------------------------------
void ControlPointToParameters(PF_InData		*in_data,
							PF_OutData		*out_data,
							PF_ParamDef *params[],
							control_point *cp,
							int iXformsOnly)
{
	int xform, var, coeff, index;

	if (!iXformsOnly)
	{
		params[AEF_CENTER_OFFSET_X]->u.fs_d.value = cp->center[0];
		params[AEF_CENTER_OFFSET_X]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_CENTER_OFFSET_Y]->u.fs_d.value = cp->center[1];
		params[AEF_CENTER_OFFSET_Y]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_PIXELS_PER_UNIT]->u.fs_d.value = cp->pixels_per_unit;
		params[AEF_PIXELS_PER_UNIT]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_SPATIAL_OVERSAMPLE]->u.sd.value = cp->spatial_oversample;
		params[AEF_SPATIAL_OVERSAMPLE]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_SPATIAL_FILTER_RADIUS]->u.fs_d.value = cp->spatial_filter_radius;
		params[AEF_SPATIAL_FILTER_RADIUS]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_SAMPLE_DENSITY]->u.fs_d.value = cp->sample_density;
		params[AEF_SAMPLE_DENSITY]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_GAMMA]->u.fs_d.value = cp->gamma;
		params[AEF_GAMMA]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_CONTRAST]->u.fs_d.value = cp->contrast;
		params[AEF_CONTRAST]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_BRIGHTNESS]->u.fs_d.value = cp->brightness;
		params[AEF_BRIGHTNESS]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_WHITE_LEVEL]->u.sd.value = cp->white_level;
		params[AEF_WHITE_LEVEL]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
	//		params[AEF_COLORMAP]->u.sd.value = cp->cmap_index;
	//		params[AEF_COLORMAP]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_COLORMAP_INTERPOLATE]->u.sd.value = cp->cmap_inter;
		params[AEF_COLORMAP_INTERPOLATE]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_NUM_BATCHES]->u.sd.value = cp->nbatches;
		params[AEF_NUM_BATCHES]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
		params[AEF_ZOOM]->u.fs_d.value = cp->zoom;
		params[AEF_ZOOM]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
	}

	for ( xform = 0 ; xform < kNumXforms ; xform++ )
	{
		if (cp->xform[xform].density > 0.0)
		{

			switch (xform)
			{
				case 0 : 
					params[AEF_XFORM0_COLOR]->u.fs_d.value = cp->xform[xform].color;
					params[AEF_XFORM0_COLOR]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					params[AEF_XFORM0_DENSITY]->u.fs_d.value = cp->xform[xform].density;
					params[AEF_XFORM0_DENSITY]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					index = AEF_XFORM0_VAR0; 
					break;
				case 1 : 
					params[AEF_XFORM1_COLOR]->u.fs_d.value = cp->xform[xform].color;
					params[AEF_XFORM1_COLOR]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					params[AEF_XFORM1_DENSITY]->u.fs_d.value = cp->xform[xform].density;
					params[AEF_XFORM1_DENSITY]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					index = AEF_XFORM1_VAR0; 
					break;
				case 2 : 
					params[AEF_XFORM2_COLOR]->u.fs_d.value = cp->xform[xform].color;
					params[AEF_XFORM2_COLOR]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					params[AEF_XFORM2_DENSITY]->u.fs_d.value = cp->xform[xform].density;
					params[AEF_XFORM2_DENSITY]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					index = AEF_XFORM2_VAR0; 
					params[AEF_XFORM2_ENABLED]->u.bd.value = TRUE;
					params[AEF_XFORM2_ENABLED]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					break;
				case 3 : 
					params[AEF_XFORM3_COLOR]->u.fs_d.value = cp->xform[xform].color;
					params[AEF_XFORM3_COLOR]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					params[AEF_XFORM3_DENSITY]->u.fs_d.value = cp->xform[xform].density;
					params[AEF_XFORM3_DENSITY]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					index = AEF_XFORM3_VAR0; 
					params[AEF_XFORM3_ENABLED]->u.bd.value = TRUE;
					params[AEF_XFORM3_ENABLED]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					break;
/*							case 4 : index = AEF_XFORM4_VAR0; break;
				case 5 : index = AEF_XFORM5_VAR0; break;
*/							default: break;
			}
			
			for ( var = 0 ; var < kNumVariations ; var++ )
			{
				params[index+var]->u.fs_d.value = cp->xform[xform].var[var] * 100.0;
				params[index+var]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
			}

			switch (xform)
			{
				case 0 : index = AEF_XFORM0_COEF0; break;
				case 1 : index = AEF_XFORM1_COEF0; break;
				case 2 : index = AEF_XFORM2_COEF0; break;
				case 3 : index = AEF_XFORM3_COEF0; break;
/*							case 4 : index = AEF_XFORM4_COEF0; break;
				case 5 : index = AEF_XFORM5_COEF0; break;
*/							default: break;
			}
			for ( coeff = 0 ; coeff < kNumCoefficients ; coeff++ )
			{
				int i, j;
				i = coeff / 2;
				j = coeff % 2;
				params[index+coeff]->u.fs_d.value = cp->xform[xform].c[i][j];
				params[index+coeff]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
			}
		}
		else
		{
			switch (xform)
			{
				case 2 : 
					params[AEF_XFORM2_ENABLED]->u.bd.value = FALSE;
					params[AEF_XFORM2_ENABLED]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					index = AEF_XFORM2_VAR0; 
					break;
				case 3 : 
					params[AEF_XFORM3_ENABLED]->u.bd.value = FALSE;
					params[AEF_XFORM3_ENABLED]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
					index = AEF_XFORM3_VAR0; 
					break;
/*							case 4 : index = AEF_XFORM4_VAR0; break;
				case 5 : index = AEF_XFORM5_VAR0; break;
*/							default: break;
			}
		}
	}
}

//=========================================================================
// 
//-------------------------------------------------------------------------
void LoadSettings(PF_InData		*in_data,
							PF_OutData		*out_data,
							PF_ParamDef *params[],
							int iSettingsNum)
{
	if (iSettingsNum == 0)
	{
		// do nothing
	}
	else
	{
		int done = FALSE;
		int lineNum = 1;
		int dest = 0;
		char *str = NULL;
		FILE *f = NULL;
		str = malloc(4000);
		if (str != NULL)
		{
			str[0] = '\0';
			f = fopen(SETTINGS_FILENAME, "r");
			if (f != NULL)
			{
				do 
				{
					char c = getc(f);
					if (c == EOF)
					{
						str[dest++] = '\0';
						done = TRUE;
					}
					else if (c == '\n')
					{
						if (lineNum == iSettingsNum)
						{
							str[dest++] = '\0';
							done = TRUE;
						}
						lineNum++;
					}
					else if (lineNum == iSettingsNum)
					{
						str[dest++] = c;
					}
				} 
				while (!done);

				if (str[0] != '\0')		// maybe there aren't enough lines in the file...
				{
					control_point cp;
					char *ss = str;

					memset((void *)&cp, 0, sizeof(control_point) );

					parse_control_point(params, &ss, &cp);
					ControlPointToParameters(in_data, out_data, params, &cp, FALSE);
				}
				else
				{
					// not enough lines in the file, do nothing
				}	
				fclose(f);
			}
			free(str);
		}
	}
}


//=========================================================================
// 
//-------------------------------------------------------------------------
void GetColorFromCMap(clrmap cmap, int cmap_len, double index, double hue_rotation, double oColorRGB[])
{
	int topIndex, botIndex, i;
	double fraction; 
	double rgb[3];
	double hsv[3];
	
	botIndex = (int)index;
	fraction = index - (double)botIndex;
	botIndex = botIndex % cmap_len;
	topIndex = (botIndex + 1) % cmap_len;		
	
	// botIndex and topIndex are integral floor(index) and ceil(index) (modulo cmap length)
	// fraction is from 0.0 to 1.0
	// fraction 0.0 means use bot, 1.0 means use top, val in between means interpolate color
	if (fraction == 0.0)
	{
		for ( i = 0 ; i < 3 ; i++ )
			rgb[i] = cmap[botIndex][i];
	}
	else if (fraction == 1.0)
	{
		for ( i = 0 ; i < 3 ; i++ )
			rgb[i] = cmap[topIndex][i];
	}
	else
	{
		// gotta interpolate
		for ( i = 0 ; i < 3 ; i++ )
			rgb[i] = (cmap[botIndex][i] + (((double)cmap[topIndex][i] - (double)cmap[botIndex][i]) * fraction));
	}

	// now apply hue rotation
	// hue_rotation should be from 0.0 (no change) to 359.99999
	while (hue_rotation >= 360.0)
		hue_rotation -= 360.0;
		
	if (hue_rotation != 0.0)
	{
		// convert to hsv
		rgb2hsv(rgb, hsv);

		// hsv[0] is hue, from 0..6, so increase it appropriately.
		hsv[0] += (hue_rotation / 60.0);

		while (hsv[0] >= 6.0)
			hsv[0] -= 6.0;

		// convert back to rgb
		hsv2rgb(hsv, rgb);
	}

	// copy to our output color
	for ( i = 0 ; i < 3 ; i++ )
		oColorRGB[i] = rgb[i];
}
//=========================================================================
// 
//-------------------------------------------------------------------------
void InvertColorMap(PF_ParamDef *params[], int n, clrmap c, int cmap_len)
{
	if (params[AEF_COLORMAP_INVERT]->u.bd.value)
	{
		int i, j;
		clrmap tmp;
		for (i = 0; i < cmap_len; i++) 
		{
			for (j = 0; j < 3; j++)
				tmp[i][j] = c[i][j];
		}

		for (i = 0; i < cmap_len; i++) 
		{
			int src = (cmap_len-1) - i;
			for (j = 0; j < 3; j++)
				c[i][j] = tmp[src][j];
		}
	}
}


//=========================================================================
// 
//-------------------------------------------------------------------------
void ModifyColorMap(PF_ParamDef *params[], int n, clrmap c, int cmap_len)
{
	InvertColorMap(params, n, c, cmap_len);

	if ((params[AEF_HUE_ROTATION]->u.fd.value == 0) 
		&& ((params[AEF_COLORMAP_ROTATION]->u.fs_d.value == 0.0) 
		|| (params[AEF_COLORMAP_ROTATION]->u.fs_d.value == 360.0)))
	{
		// do nothing... 
		// maybe someday we'll do something here...
	}
	else
	{
		int i, j;
		clrmap tmp;
		for (i = 0; i < cmap_len; i++) 
		{
			for (j = 0; j < 3; j++)
				tmp[i][j] = c[i][j];
		}

		for (i = 0; i < cmap_len; i++) 
		{
			PF_Fixed hueRotation = params[AEF_HUE_ROTATION]->u.fd.value;
			double src = i + params[AEF_COLORMAP_ROTATION]->u.fs_d.value;
			for (j = 0; j < 3; j++)
				GetColorFromCMap(tmp, cmap_len, src, FIX2FLOAT(hueRotation), c[i]);
				//c[i][j] = tmp[src][j];
		}
	}
}





//=========================================================================
// try to convince AE to change the slider name.
//-------------------------------------------------------------------------
void ChangeSliderName(	PF_InData		*in_data,
						PF_OutData		*out_data,
						PF_ParamDef		*params[],
						int iParamNum,
						PF_ParamFlags flags,
						PF_ParamUIFlags ui_flags,
						char *iNewName) 
{
#if 0
	PF_ParamUtilsSuite1	*pus	= NULL;
	PF_Err				err		= AEFX_AcquireSuite(in_data, 
													out_data, 
													kPFParamUtilsSuite, 
													kPFParamUtilsSuiteVersion1, 
													"Couldn't load suite.",
													&pus);
	if (!err && pus && iNewName) 
	{
		PF_ParamDef	def;
/*		ui_flags: PF_PUI_ECW_SEPARATOR, PF_PUI_DISABLED only
		ui_width
		ui_height
		name
		flags: PF_ParamFlag_COLLAPSE_TWIRLY only

		PF_ParamDefUnion:
			slider_min, slider_max, precision, display_flags of any slider type
*/
		AEFX_CLR_STRUCT(def);	
		memcpy(&def, params[iParamNum], sizeof(PF_ParamDef));
		def.flags = flags;
		def.ui_flags = ui_flags;
		PF_STRCPY(def.name, iNewName); 
		err =  (pus->PF_UpdateParamUI)(	in_data->effect_ref, iParamNum,	&def);
	}
	AEFX_ReleaseSuite(in_data, out_data, kPFParamUtilsSuite, kPFParamUtilsSuiteVersion1, "Couldn't release suite.");
#endif
}



//=========================================================================
// 
//-------------------------------------------------------------------------
#define SetParam(p, ind, val) \
	do { p[ind]->u.sd.value = val; \
	p[ind]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE; } while (0)
#define SetFloatParam(p, ind, val) \
	do { p[ind]->u.fs_d.value = val; \
	p[ind]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE; } while (0)
#define SetFloatParamDelta(p, ind, val) \
	do { p[ind]->u.fs_d.value = val + (double)(p[ind]->u.fs_d.value); \
	p[ind]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE; } while (0)
/*		// doesn't work dammit
#define SetParamName(p, ind, newname) \
	do { PF_STRCPY(p[ind]->name, newname); \
	p[ind]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE; } while (0)
*/
//=========================================================================
// 
//-------------------------------------------------------------------------
PF_Err	HandleChangedParam(	PF_InData		*in_data,
							PF_OutData		*out_data,
							PF_ParamDef		*params[],
							PF_LayerDef		*output,
							PF_UserChangedParamExtra	*extra)
{
	PF_Err			err			= PF_Err_NONE;
	
	// what does this do???
/*	if (in_data->sequence_data == NULL)	
	{
		err = SequenceResetup(in_data, out_data, params, output);
	}
*/

	switch (extra->param_index) 
	{
		case AEF_COMMAND_MENU:
			srandom(time(0));
			HandleMenuCommand(in_data, out_data, params, output, extra, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
		case AEF_RANDOMIZE_COMMAND_MENU:
			srandom(time(0));
			HandleRandomizeMenuCommand(in_data, out_data, params, output, extra, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
		case AEF_XFORM0_COMMAND_MENU:
			srandom(time(0));
			HandleXformMenuCommand(in_data, out_data, params, output, extra, 0, AEF_XFORM0_VAR0, AEF_XFORM0_COEF0, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
		case AEF_XFORM1_COMMAND_MENU:
			srandom(time(0));
			HandleXformMenuCommand(in_data, out_data, params, output, extra, 1, AEF_XFORM1_VAR0, AEF_XFORM1_COEF0, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
		case AEF_XFORM2_COMMAND_MENU:
			srandom(time(0));
			HandleXformMenuCommand(in_data, out_data, params, output, extra, 2, AEF_XFORM2_VAR0, AEF_XFORM2_COEF0, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
		case AEF_XFORM3_COMMAND_MENU:
			srandom(time(0));
			HandleXformMenuCommand(in_data, out_data, params, output, extra, 3, AEF_XFORM3_VAR0, AEF_XFORM3_COEF0, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
/*		case AEF_XFORM4_COMMAND_MENU:
			srandom(time(0));
			HandleXformMenuCommand(in_data, out_data, params, output, extra, 4, AEF_XFORM4_VAR0, AEF_XFORM4_COEF0, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
		case AEF_XFORM5_COMMAND_MENU:
			srandom(time(0));
			HandleXformMenuCommand(in_data, out_data, params, output, extra, 5, AEF_XFORM5_VAR0, AEF_XFORM5_COEF0, extra->param_index, params[extra->param_index]->u.pd.value);
			break;
*/		case AEF_LOAD_PARAMETERS : 
			LoadSettings(in_data, out_data, params, params[extra->param_index]->u.sd.value);
			break;

		// doesn't work dammit... why doesn't this update while the user is dragging the slider??
/*		case AEF_COLORMAP:
			{
				char str[256];
				
				if (params[extra->param_index]->u.sd.value < NumCMaps())
					sprintf(str, "Colormap %ld: %s", params[extra->param_index]->u.sd.value, cmapNames[params[extra->param_index]->u.sd.value]);
				else
				{
					char fName[MAX_PATH];
					long fNum = params[extra->param_index]->u.sd.value - NumCMaps();
					GetNthFile(CMAP_DIRECTORY, "*.MAP", fNum, fName);
					RemoveFullPath(fName);
					sprintf(str, "Colormap %ld (%ld): %s", params[extra->param_index]->u.sd.value, fNum, fName);
				}
				// change the slider name dynamically 
				ChangeSliderName(in_data, out_data, params, extra->param_index, PF_ParamFlag_SUPERVISE, 0, str);
			}
			break;
*/		default:
			break;
	} 

	return err;
}


//=========================================================================
// menu command items
//-------------------------------------------------------------------------

enum {
	kCmdChoose = 1,
	kSep0,
	kCmdSaveSettings,					// saves settings in SETTINGS_FILENAME
	kSep1,
	kCmdAllXformsLinearOnly,			// sets all xforms to these variations (i.e. 100% this one, 0% everything else)
	kCmdAllXformsSinusoidalOnly,
	kCmdAllXformsSphericalOnly,
	kCmdAllXformsSwirlOnly,
	kCmdAllXformsHorseshoeOnly,
	kCmdAllXformsPolarOnly,
	kCmdAllXformsBentOnly,
	kSep2,
	kCmdDefaultView,					// default zoom/re-center/etc.
	kCmdDefaultColorCorrection,			// reset color correction
	kSep3,
	kCmdDecreaseFOV,					// pixels per unit up/down
	kCmdIncreaseFOV,
	kCmdZoomIn,							// zoom up/down
	kCmdZoomOut,
//	kCmdUseFOVNotZoom,
//	kCmdUseZoomNotFOV,
	kSep4,
	kCmdFastestQuality,					// set sample density/spatial oversample/etc to good values
	kCmdDraftQuality,
	kCmdMediumQuality,
	kCmdHighQuality,
	kCmdVeryHighQuality,
	kCmdUltraHighQuality,
	kSep5,
	kCmdInsertKeys,						// insert keys in all tracks w/function curves activated
	kSep6,
	kCmdAllVars0,						// set all xform variations to these values
	kCmdAllVars100,
	kCmdAllCoeffsMinus1,				// set all xform coefficients to these values
	kCmdAllCoeffs0,
	kCmdAllCoeffs1,
	kSep7,
	kCmdGNU,
	kNumMenuCommands
};


enum {
	kCmdRndChoose = 1,
	kRndSep0,
	kCmdRndRandomColormap,
	kRndSep1,
	kCmdRndRandomizeAll,
	kCmdRndRandomizeAllSmart,
	kRndSep1a,
	kCmdRndRandomizeVarsVectorForward,
	kCmdRndRandomizeVarsVectorBackward,
	kCmdRndRandomizeCoeffsVectorForward,
	kCmdRndRandomizeCoeffsVectorBackward,
	kCmdRndRandomizeVectorForward,
	kCmdRndRandomizeVectorBackward,
	kRndSep2,
	kCmdRndRandomizeVarsSlightly,
	kCmdRndRandomizeVarsALittle,
	kCmdRndRandomizeVarsSome,
	kCmdRndRandomizeVarsLots,
	kCmdRndRandomizeVarsTotally,
	kRndSep3,
	kCmdRndRandomizeCoeffsSlightly,
	kCmdRndRandomizeCoeffsALittle,
	kCmdRndRandomizeCoeffsSome,
	kCmdRndRandomizeCoeffsLots,
	kCmdRndRandomizeCoeffsTotally,
	kRndSep4,
	kCmdRndRandomParameterSet,
	kNumRandomizeMenuCommands
};
#define CMD_MENU_ITEMS \
	"Choose...|" \
	"-(|" \
	"Save Settings|" \
	"-(|" \
	"All Xforms Linear Only|" \
	"All Xforms Sinusoidal Only|" \
	"All Xforms Spherical Only|" \
	"All Xforms Swirl Only|" \
	"All Xforms Horseshoe Only|" \
	"All Xforms Polar Only|" \
	"All Xforms Bent Only|" \
	"-(|" \
	"Default View|" \
	"Default Color Correction|" \
	"-(|" \
	"Narrow FOV|" \
	"Widen FOV|" \
	"Zoom In|" \
	"Zoom Out|" \
/*	"Use FOV not Zoom|" \
	"Use Zoom not FOV|" \
*/	"-(|" \
	"Fastest Quality|" \
	"Draft Quality|" \
	"Medium Quality|" \
	"High Quality|" \
	"Very High Quality|" \
	"Ultra High Quality|" \
	"-(|" \
	"Insert Keys|" \
	"-(|" \
	"Set All Xform Vars = 0.0|" \
	"Set All Xform Vars = 100.0|" \
	"Set All Xform Coeffs = minus 1.0|" \
	"Set All Xform Coeffs = 0.0|" \
	"Set All Xform Coeffs = 1.0|" \
	"-(|" \
	"View GNU License"

#define RANDOMIZE_CMD_MENU_ITEMS \
	"Global Randomize...|" \
	"-(|" \
	"Random Color Map|" \
	"-(|" \
	"Randomize All|" \
	"Randomize All Smart|" \
	"-(|" \
	"Vector All Vars Forward|" \
	"Vector All Vars Backward|" \
	"Vector All Coeffs Forward|" \
	"Vector All Coeffs Backward|" \
	"Vector All Forward|" \
	"Vector All Backward|" \
	"-(|" \
	"Randomize All Vars Slightly|" \
	"Randomize All Vars A Little|" \
	"Randomize All Vars Some|" \
	"Randomize All Vars Lots|" \
	"Randomize All Vars Totally|" \
	"-(|" \
	"Randomize All Coeffs Slightly|" \
	"Randomize All Coeffs A Little|" \
	"Randomize All Coeffs Some|" \
	"Randomize All Coeffs Lots|" \
	"Randomize All Coeffs Totally|" \
	"-(|" \
	"Random Parameter Set" 

enum {
	kCmdXformChoose = 1,
	kXformSep1,
	kCmdXformVarLinearOnly,
	kCmdXformVarSinusoidalOnly,
	kCmdXformVarSphericalOnly,
	kCmdXformVarSwirlOnly,
	kCmdXformVarHorseshoeOnly,
	kCmdXformVarPolarOnly,
	kCmdXformVarBentOnly,
	kXformSep1a,
	kCmdXformRandomizeVarsVectorForward,
	kCmdXformRandomizeVarsVectorBackward,
	kCmdXformRandomizeCoeffsVectorForward,
	kCmdXformRandomizeCoeffsVectorBackward,
	kCmdXformRandomizeVectorForward,
	kCmdXformRandomizeVectorBackward,
	kXformSep2,
	kCmdXformRandomizeVarsSlightly,
	kCmdXformRandomizeVarsALittle,
	kCmdXformRandomizeVarsSome,
	kCmdXformRandomizeVarsLots,
	kCmdXformRandomizeVarsTotally,
	kXformSep3,
	kCmdXformRandomizeCoeffsSlightly,
	kCmdXformRandomizeCoeffsALittle,
	kCmdXformRandomizeCoeffsSome,
	kCmdXformRandomizeCoeffsLots,
	kCmdXformRandomizeCoeffsTotally,
	kXformSep4,
	kCmdXformRandomizeAll,
	kXformSep5,
	kCmdXformInsertKeys,
	kXformSep6,
	kCmdXformAllVars0,
	kCmdXformAllVars100,
	kCmdXformAllCoeffsMinus1,
	kCmdXformAllCoeffs0,
	kCmdXformAllCoeffs1,
	kNumXformMenuCommands
};

#define XFORM_CMD_MENU_ITEMS \
	"-(|" \
	"Linear Only|" \
	"Sinusoidal Only|" \
	"Spherical Only|" \
	"Swirl Only|" \
	"Horseshoe Only|" \
	"Polar Only|" \
	"Bent Only|" \
	"-(|" \
	"Vector Vars Forward|" \
	"Vector Vars Backward|" \
	"Vector Coeffs Forward|" \
	"Vector Coeffs Backward|" \
	"Vector Forward|" \
	"Vector Backward|" \
	"-(|" \
	"Randomize Vars Slightly|" \
	"Randomize Vars A Little|" \
	"Randomize Vars Some|" \
	"Randomize Vars Lots|" \
	"Randomize Vars Totally|" \
	"-(|" \
	"Randomize Coeffs Slightly|" \
	"Randomize Coeffs A Little|" \
	"Randomize Coeffs Some|" \
	"Randomize Coeffs Lots|" \
	"Randomize Coeffs Totally|" \
	"-(|" \
	"Randomize This Xform Totally|" \
	"-(|" \
	"Insert Keys for This Xform|" \
	"-(|" \
	"Set All Vars = 0.0|" \
	"Set All Vars = 100.0|" \
	"Set All Coeffs = minus 1.0|" \
	"Set All Coeffs = 0.0|" \
	"Set All Coeffs = 1.0" 

#define XFORM0_CMD_MENU_ITEMS "Xform 0 Choose...|" XFORM_CMD_MENU_ITEMS
#define XFORM1_CMD_MENU_ITEMS "Xform 1 Choose...|" XFORM_CMD_MENU_ITEMS
#define XFORM2_CMD_MENU_ITEMS "Xform 2 Choose...|" XFORM_CMD_MENU_ITEMS
#define XFORM3_CMD_MENU_ITEMS "Xform 3 Choose...|" XFORM_CMD_MENU_ITEMS
#define XFORM4_CMD_MENU_ITEMS "Xform 4 Choose...|" XFORM_CMD_MENU_ITEMS
#define XFORM5_CMD_MENU_ITEMS "Xform 5 Choose...|" XFORM_CMD_MENU_ITEMS

// amount to vary for slightly/a little/some/lots
// values pulled out of my ass.
// variations amounts must be from 0..100
// coeffs amounts must be from 0..1
// these should be changed to a slider maybe(?) 
// (i.e. use the slider to set the randomize amount, then choose "randomize")
#define AMOUNT_VAR_SLIGHTLY		(0.03 * 100.0)
#define AMOUNT_VAR_A_LITTLE		(0.10 * 100.0)
#define AMOUNT_VAR_SOME			(0.20 * 100.0)
#define AMOUNT_VAR_LOTS			(0.3 * 100.0)
#define AMOUNT_COEFF_SLIGHTLY   0.03
#define AMOUNT_COEFF_A_LITTLE	0.10
#define AMOUNT_COEFF_SOME		0.20
#define AMOUNT_COEFF_LOTS		0.3


//=========================================================================
// 
//-------------------------------------------------------------------------
void DoCmdForAllXforms(PF_InData		*in_data,
						PF_OutData		*out_data,
						PF_ParamDef		*params[],
						PF_LayerDef		*output,
						PF_UserChangedParamExtra	*extra,
						int iParamIndex,
						int cmd)
{
	HandleXformMenuCommand(in_data, out_data, params, output, extra, 0, AEF_XFORM0_VAR0, AEF_XFORM0_COEF0, iParamIndex, cmd);
	HandleXformMenuCommand(in_data, out_data, params, output, extra, 1, AEF_XFORM1_VAR0, AEF_XFORM1_COEF0, iParamIndex, cmd);
	HandleXformMenuCommand(in_data, out_data, params, output, extra, 2, AEF_XFORM2_VAR0, AEF_XFORM2_COEF0, iParamIndex, cmd);
	HandleXformMenuCommand(in_data, out_data, params, output, extra, 3, AEF_XFORM3_VAR0, AEF_XFORM3_COEF0, iParamIndex, cmd);
/*	HandleXformMenuCommand(in_data, out_data, params, output, extra, 4, AEF_XFORM4_VAR0, AEF_XFORM4_COEF0, iParamIndex, cmd);
	HandleXformMenuCommand(in_data, out_data, params, output, extra, 5, AEF_XFORM5_VAR0, AEF_XFORM5_COEF0, iParamIndex, cmd);
*/
}

//=========================================================================
// 
//-------------------------------------------------------------------------
void HandleMenuCommand(PF_InData		*in_data,
						PF_OutData		*out_data,
						PF_ParamDef		*params[],
						PF_LayerDef		*output,
						PF_UserChangedParamExtra	*extra,
						int iParamIndex,
						int iMenuCommand)
{
	int cmd;

	params[iParamIndex]->u.sd.value = kCmdChoose;		// make menu go back to choose
	params[iParamIndex]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;

	switch (iMenuCommand)
	{
		case kCmdChoose:
			// ignore this command entirely
			break;

		case kCmdSaveSettings:
			{
				FILE *f = NULL;
				char *str = malloc(8000);

				if (str != NULL)
				{
					char str0[256];
					str[0] = '\0';
					AppendGlobalParamsString(in_data, params, str, 0);
					sprintf(str0, "time 0 cmap %ld ", params[AEF_COLORMAP]->u.sd.value);
					strcat(str, str0);
					AppendXform0String(params, str);
					AppendXform1String(params, str);
					if (params[AEF_XFORM2_ENABLED]->u.bd.value)
						AppendXform2String(params, str);
					if (params[AEF_XFORM3_ENABLED]->u.bd.value)
						AppendXform3String(params, str);
					strcat(str, ";\n");
					f = fopen(SETTINGS_FILENAME, "a+");
					if (f != NULL)
					{
						fwrite(str, sizeof(char), strlen(str), f);
						fclose(f);
					}
					free(str);

					// ABDTODO this would be cool
					// set the load parameters slider to these settings
				//	params[AEF_LOAD_PARAMETERS]->u.sd.value = GetNumLines();
				//	params[AEF_LOAD_PARAMETERS]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;

					// let the user know what number this is
/*					PF_SPRINTF(str0, "settings saved in slot %d", 0);
					PF_STRCPY(out_data->return_msg, str0);
*/					
					PF_SPRINTF(str0, "Parameter Set (of %ld)", GetNumLines(SETTINGS_FILENAME));
					ChangeSliderName(in_data, out_data, params, AEF_LOAD_PARAMETERS, PF_ParamFlag_SUPERVISE | PF_ParamFlag_CANNOT_TIME_VARY, 0, str0);
				}
			}
			break;

		case kCmdDefaultView:
			SetFloatParam(params, AEF_CENTER_OFFSET_X, 0.0);
			SetFloatParam(params, AEF_CENTER_OFFSET_Y, 0.0);
		//	SetFixedParam(params, AEF_ALPHA_BLEND, BLEND_DFLT);
		//	SetFixedParam(params, AEF_BLEND, BLEND_DFLT);
			SetFloatParam(params, AEF_PIXELS_PER_UNIT, 100.0);
			SetFloatParam(params, AEF_ZOOM, 0.0);
			break;

		case kCmdDefaultColorCorrection:
			SetFloatParam(params, AEF_GAMMA, 2.0);
			SetFloatParam(params, AEF_CONTRAST, 1.0);
			SetFloatParam(params, AEF_BRIGHTNESS, 1.5);
			SetParam(params, AEF_WHITE_LEVEL, 200);
			break;

		case kCmdFastestQuality:
			// the absolute fastest quality you can get and still have a picture of sorts
			SetFloatParam(params, AEF_SAMPLE_DENSITY, 0.01);
			SetParam(params, AEF_SPATIAL_OVERSAMPLE, 1);
			SetFloatParam(params, AEF_SPATIAL_FILTER_RADIUS, 0.0);
			// ABDTODO probably should swap out Zoom and FOV here...
			// (need to have Zoom 0.0 for max speed...)
			break;
		case kCmdDraftQuality:
			// sample density should be spatial oversample squared to 
			// have the same approximate density for various quality settings
			SetFloatParam(params, AEF_SAMPLE_DENSITY, 1.0);
			SetParam(params, AEF_SPATIAL_OVERSAMPLE, 1);
			SetFloatParam(params, AEF_SPATIAL_FILTER_RADIUS, 0.0);
			// ABDTODO probably should swap out Zoom and FOV here...
			break;
		case kCmdMediumQuality:
			// sample density should be spatial oversample squared to 
			// have the same approximate density for various quality settings
			SetFloatParam(params, AEF_SAMPLE_DENSITY, 4.0);
			SetParam(params, AEF_SPATIAL_OVERSAMPLE, 2);
			SetFloatParam(params, AEF_SPATIAL_FILTER_RADIUS, 0.5);
			// ABDTODO probably should swap out Zoom and FOV here...
			break;
		case kCmdHighQuality:
			// sample density should be spatial oversample squared to 
			// have the same approximate density for various quality settings
			SetFloatParam(params, AEF_SAMPLE_DENSITY, 16.0);
			SetParam(params, AEF_SPATIAL_OVERSAMPLE, 4);
			SetFloatParam(params, AEF_SPATIAL_FILTER_RADIUS, 0.5);
			// ABDTODO probably should swap out Zoom and FOV here...
			break;
		case kCmdVeryHighQuality:
			// sample density should be spatial oversample squared to 
			// have the same approximate density for various quality settings
			SetFloatParam(params, AEF_SAMPLE_DENSITY, 64.0);
			SetParam(params, AEF_SPATIAL_OVERSAMPLE, 8);
			SetFloatParam(params, AEF_SPATIAL_FILTER_RADIUS, 0.5);
			// ABDTODO probably should swap out Zoom and FOV here...
			break;
		case kCmdUltraHighQuality:
			// sample density should be spatial oversample squared to 
			// have the same approximate density for various quality settings
			SetFloatParam(params, AEF_SAMPLE_DENSITY, 100.0);
			SetParam(params, AEF_SPATIAL_OVERSAMPLE, 10);
			SetFloatParam(params, AEF_SPATIAL_FILTER_RADIUS, 0.5);
			// ABDTODO probably should swap out Zoom and FOV here...
			break;
		case kCmdDecreaseFOV:
			{
				double val = (double)params[AEF_PIXELS_PER_UNIT]->u.fs_d.value * 1.3;
				SetFloatParam(params, AEF_PIXELS_PER_UNIT, val);
			}
			break;
		case kCmdIncreaseFOV:
			{
				double val = (double)params[AEF_PIXELS_PER_UNIT]->u.fs_d.value * 0.77;
				SetFloatParam(params, AEF_PIXELS_PER_UNIT, val);
			}
			break;
		case kCmdZoomIn:
			{
				double val;
				if ((double)params[AEF_ZOOM]->u.fs_d.value == 0.0)
					val = 0.1;
				else
					val = (double)params[AEF_ZOOM]->u.fs_d.value * 1.3;
				SetFloatParam(params, AEF_ZOOM, val);
			}
			break;
		case kCmdZoomOut:
			{
				double val = (double)params[AEF_ZOOM]->u.fs_d.value * 0.77;
				SetFloatParam(params, AEF_ZOOM, val);
			}
			break;
/*		case kCmdUseFOVNotZoom:
			{
				double val = (double)params[AEF_PIXELS_PER_UNIT]->u.fs_d.value * ((double)params[AEF_ZOOM]->u.fs_d.value + 1.0);
				SetFloatParam(params, AEF_PIXELS_PER_UNIT, val);
				SetFloatParam(params, AEF_ZOOM, 0.0);
			}
			break;
		case kCmdUseZoomNotFOV:
			{
				double val = ((double)params[AEF_PIXELS_PER_UNIT]->u.fs_d.value - 100.0) / 100.0;
				SetFloatParam(params, AEF_PIXELS_PER_UNIT, 100.0);
				if (val > 0.000001)
					SetFloatParam(params, AEF_ZOOM, val);
			}
			break;
*/
		case kCmdAllXformsLinearOnly:
			cmd = kCmdXformVarLinearOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllXformsSinusoidalOnly:
			cmd = kCmdXformVarSinusoidalOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllXformsSphericalOnly:
			cmd = kCmdXformVarSphericalOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllXformsSwirlOnly:
			cmd = kCmdXformVarLinearOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllXformsHorseshoeOnly:
			cmd = kCmdXformVarHorseshoeOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllXformsPolarOnly:
			cmd = kCmdXformVarPolarOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllXformsBentOnly:
			cmd = kCmdXformVarBentOnly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;

		case kCmdInsertKeys:
			{
				int index;
				// pretend like we changed all the parameter values (to fake AE into inserting keys)
				for ( index = AEF_CENTER_OFFSET_X ; index < AEF_RANDOM_SEED ; index++ )
					params[index]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;

				cmd = kCmdXformInsertKeys;
				DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			}
			break;

		case kCmdAllVars0:
			cmd = kCmdXformAllVars0;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllVars100:
			cmd = kCmdXformAllVars100;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllCoeffsMinus1:
			cmd = kCmdXformAllCoeffsMinus1;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllCoeffs0:
			cmd = kCmdXformAllCoeffs0;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdAllCoeffs1:
			cmd = kCmdXformAllCoeffs1;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
/*		case kCmdGNU:
			PF_STRCPY(	out_data->return_msg, 
						"This software is distributed under the GNU General Public License. "
						"ABSOLUTELY NO WARRANTY. See COPYING.txt for full details."
						"If you did not receive a copy of that file "
						"with this software write to Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA"
						);
			break;
*/		default:
			break;
	}
}

//=========================================================================
// 
//-------------------------------------------------------------------------
void HandleRandomizeMenuCommand(PF_InData		*in_data,
						PF_OutData		*out_data,
						PF_ParamDef		*params[],
						PF_LayerDef		*output,
						PF_UserChangedParamExtra	*extra,
						int iParamIndex,
						int iMenuCommand)
{
	int cmd;
	extern int gNumCmaps;

	params[iParamIndex]->u.sd.value = kCmdChoose;		// make menu go back to choose
	params[iParamIndex]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;

	switch (iMenuCommand)
	{
		case kCmdChoose:
			// ignore this command entirely
			break;

		case kCmdRndRandomColormap:
			{
				int cmap;
				do
				{
					cmap = random() % gNumCmaps;
				}
				while (cmap == params[AEF_COLORMAP]->u.sd.value);
				SetParam(params, AEF_COLORMAP, cmap);
			}
			break;
		case kCmdRndRandomParameterSet:
			{
				int numLines = GetNumLines(SETTINGS_FILENAME);
				int settingsNum;
				do
				{
					settingsNum = random() % numLines;
				}
				while (settingsNum == params[AEF_LOAD_PARAMETERS]->u.sd.value);
				LoadSettings(in_data, out_data, params, settingsNum);
				params[AEF_LOAD_PARAMETERS]->u.sd.value = settingsNum;
				params[AEF_LOAD_PARAMETERS]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
			}
			break;
		case kCmdRndRandomizeVarsSlightly:
			cmd = kCmdXformRandomizeVarsSlightly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeVarsALittle:
			cmd = kCmdXformRandomizeVarsALittle;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeVarsSome:
			cmd = kCmdXformRandomizeVarsSome;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeVarsLots:
			cmd = kCmdXformRandomizeVarsLots;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeVarsTotally:
			cmd = kCmdXformRandomizeVarsTotally;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeCoeffsSlightly:
			cmd = kCmdXformRandomizeCoeffsSlightly;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeCoeffsALittle:
			cmd = kCmdXformRandomizeCoeffsALittle;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeCoeffsSome:
			cmd = kCmdXformRandomizeCoeffsSome;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeCoeffsLots:
			cmd = kCmdXformRandomizeCoeffsLots;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeCoeffsTotally:
			cmd = kCmdXformRandomizeCoeffsTotally;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeAll:
			cmd = kCmdXformRandomizeAll;
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, cmd);
			break;
		case kCmdRndRandomizeAllSmart:
			{
				control_point cp;
				random_control_point(params, &cp, variation_random);
				ControlPointToParameters(in_data, out_data, params, &cp, TRUE);
			}
			break;
		case kCmdRndRandomizeVarsVectorForward:
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, kCmdXformRandomizeVarsVectorForward);
			break;
		case kCmdRndRandomizeVarsVectorBackward:
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, kCmdXformRandomizeVarsVectorBackward);
			break;
		case kCmdRndRandomizeCoeffsVectorForward:
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, kCmdXformRandomizeCoeffsVectorForward);
			break;
		case kCmdRndRandomizeCoeffsVectorBackward:
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, kCmdXformRandomizeCoeffsVectorBackward);
			break;
		case kCmdRndRandomizeVectorForward:
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, kCmdXformRandomizeVectorForward);
			break;
		case kCmdRndRandomizeVectorBackward:
			DoCmdForAllXforms(in_data, out_data, params, output, extra, iParamIndex, kCmdXformRandomizeVectorBackward);
			break;
		default:
			break;
	}
}



//=========================================================================
// 
//-------------------------------------------------------------------------
void HandleXformMenuCommand(PF_InData		*in_data,
						PF_OutData		*out_data,
						PF_ParamDef		*params[],
						PF_LayerDef		*output,
						PF_UserChangedParamExtra	*extra,
						int iXformNum,
						int iXformVarNum,
						int iXformCoeffNum,
						int iParamIndex,
						int iMenuCommand)
{
	int index;

	params[iParamIndex]->u.sd.value = kCmdXformChoose;		// make menu go back to choose
	params[iParamIndex]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;

	switch (iMenuCommand)
	{
		case kCmdXformChoose:
			// ignore this command entirely
			break;

		case kCmdXformVarLinearOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0); 
			SetFloatParam(params, iXformVarNum + 0, 100.0);
			break;

		case kCmdXformVarSinusoidalOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0); 
			SetFloatParam(params, iXformVarNum + 1, 100.0);
			break;

		case kCmdXformVarSphericalOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0); 
			SetFloatParam(params, iXformVarNum + 2, 100.0);
			break;

		case kCmdXformVarSwirlOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0); 
			SetFloatParam(params, iXformVarNum + 3, 100.0);
			break;

		case kCmdXformVarHorseshoeOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0); 
			SetFloatParam(params, iXformVarNum + 4, 100.0);
			break;

		case kCmdXformVarPolarOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0); 
			SetFloatParam(params, iXformVarNum + 5, 100.0);
			break;

		case kCmdXformVarBentOnly:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0);
			SetFloatParam(params, iXformVarNum + 6, 100.0);
			break;

		case kCmdXformRandomizeVarsVectorForward:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParamDelta(params, iXformVarNum + index, gExploreVectorVars[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			break;
		case kCmdXformRandomizeVarsVectorBackward:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParamDelta(params, iXformVarNum + index, -gExploreVectorVars[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			break;
		case kCmdXformRandomizeCoeffsVectorForward:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParamDelta(params, iXformCoeffNum + index, gExploreVectorCoeffs[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			break;
		case kCmdXformRandomizeCoeffsVectorBackward:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParamDelta(params, iXformCoeffNum + index, -gExploreVectorCoeffs[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			break;
		case kCmdXformRandomizeVectorForward:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParamDelta(params, iXformVarNum + index, gExploreVectorVars[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParamDelta(params, iXformCoeffNum + index, gExploreVectorCoeffs[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			break;
		case kCmdXformRandomizeVectorBackward:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParamDelta(params, iXformVarNum + index, -gExploreVectorVars[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParamDelta(params, iXformCoeffNum + index, -gExploreVectorCoeffs[iXformNum][index] * (double)params[AEF_RANDOM_VECTOR_LENGTH]->u.fs_d.value);
			break;

		case kCmdXformRandomizeVarsSlightly:
			for ( index = 0 ; index < kNumVariations ; index++ )
			{
				gExploreVectorVars[iXformNum][index] = frandom() * AMOUNT_VAR_SLIGHTLY;
				SetFloatParamDelta(params, iXformVarNum + index, gExploreVectorVars[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeVarsALittle:
			for ( index = 0 ; index < kNumVariations ; index++ )
			{
				gExploreVectorVars[iXformNum][index] = frandom() * AMOUNT_VAR_A_LITTLE;
				SetFloatParamDelta(params, iXformVarNum + index, gExploreVectorVars[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeVarsSome:
			for ( index = 0 ; index < kNumVariations ; index++ )
			{
				gExploreVectorVars[iXformNum][index] = frandom() * AMOUNT_VAR_SOME;
				SetFloatParamDelta(params, iXformVarNum + index, gExploreVectorVars[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeVarsLots:
			for ( index = 0 ; index < kNumVariations ; index++ )
			{
				gExploreVectorVars[iXformNum][index] = frandom() * AMOUNT_VAR_LOTS;
				SetFloatParamDelta(params, iXformVarNum + index, gExploreVectorVars[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeVarsTotally:
			for ( index = 0 ; index < kNumVariations ; index++ )
			{
				SetFloatParam(params, iXformVarNum + index, frandom() * 100.0);
			}
			break;

		case kCmdXformRandomizeCoeffsSlightly:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
			{
				gExploreVectorCoeffs[iXformNum][index] = frandom() * AMOUNT_COEFF_SLIGHTLY;
				SetFloatParamDelta(params, iXformCoeffNum + index, gExploreVectorCoeffs[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeCoeffsALittle:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
			{
				gExploreVectorCoeffs[iXformNum][index] = frandom() * AMOUNT_COEFF_A_LITTLE;
				SetFloatParamDelta(params, iXformCoeffNum + index, gExploreVectorCoeffs[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeCoeffsSome:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
			{
				gExploreVectorCoeffs[iXformNum][index] = frandom() * AMOUNT_COEFF_SOME;
				SetFloatParamDelta(params, iXformCoeffNum + index, gExploreVectorCoeffs[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeCoeffsLots:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
			{
				gExploreVectorCoeffs[iXformNum][index] = frandom() * AMOUNT_COEFF_LOTS;
				SetFloatParamDelta(params, iXformCoeffNum + index, gExploreVectorCoeffs[iXformNum][index]);
			}
			break;

		case kCmdXformRandomizeCoeffsTotally:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParam(params, iXformCoeffNum + index, frandom());
			break;

		case kCmdXformRandomizeAll:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, frandom() * 100.0);
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParam(params, iXformCoeffNum + index, frandom());
			break;

		case kCmdXformInsertKeys:
			{
				int start, end;
				switch (iXformNum)
				{
					case 0: start = AEF_XFORM0_START+1; end = AEF_XFORM0_END; break;
					case 1: start = AEF_XFORM1_START+1; end = AEF_XFORM1_END; break;
					case 2: start = AEF_XFORM2_START+1; end = AEF_XFORM2_END; break;
					case 3: start = AEF_XFORM3_START+1; end = AEF_XFORM3_END; break;
/*					case 4: start = AEF_XFORM4_START+1; end = AEF_XFORM4_END; break;
					case 5: start = AEF_XFORM5_START+1; end = AEF_XFORM5_END; break;
*/				}
				// pretend like we changed all the parameter values (to fake AE into inserting keys)
				for ( index = start ; index < end ; index++ )
					params[index]->uu.change_flags = PF_ChangeFlag_CHANGED_VALUE;
			}
			break;

		case kCmdXformAllVars0:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 0.0);
			break;

		case kCmdXformAllVars100:
			for ( index = 0 ; index < kNumVariations ; index++ )
				SetFloatParam(params, iXformVarNum + index, 100.0); 
			break;

		case kCmdXformAllCoeffsMinus1:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParam(params, iXformCoeffNum + index, -1.0);
			break;

		case kCmdXformAllCoeffs0:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParam(params, iXformCoeffNum + index, 0.0); 
			break;

		case kCmdXformAllCoeffs1:
			for ( index = 0 ; index < kNumCoefficients ; index++ )
				SetFloatParam(params, iXformCoeffNum + index, 1.0); 
			break;

		default:
			{
				// error, unknown menu command
			}
			break;
	}

/*	for ( index = AEF_XFORM0_VAR0 ; index < kNumVariations ; index++ )
		CLAMP(params[index]->u.fs_d.value, 0.0, 100.0);
	for ( index = AEF_XFORM0_COEF0 ; index < kNumCoefficients ; index++ )
		CLAMP(params[index]->u.fs_d.value, -1.0, 1.0);
*/}

//=========================================================================
// 
//-------------------------------------------------------------------------
PF_Err ParamsSetup (
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output)
{
	PF_Err			err = PF_Err_NONE;
	PF_ParamDef		def;	
	char name[256];
	long xformNum, varNum, coeffNum;

	//-------------------------------------
	// PARAMETER 0 is always the default input layer...
	// (we don't need to set that one up.)

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_COMMAND_MENU; 
	def.flags		= PF_ParamFlag_SUPERVISE | PF_ParamFlag_CANNOT_TIME_VARY;
	// can't use PF_PUI_STD_CONTROL_ONLY for the menu because AE won't handle
	// the undo stack automatically for us (?i think?)
//	def.ui_flags	= PF_PUI_STD_CONTROL_ONLY;
	def.param_type	= PF_Param_POPUP;
	def.u.pd.dephault = def.u.pd.value = kCmdChoose;
	def.u.pd.num_choices = kNumMenuCommands;
	def.u.pd.u.namesptr = CMD_MENU_ITEMS;		// just a pointer!  don't strcpy into this!
	PF_STRCPY(def.name, "Global Commands");
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_RANDOMIZE_COMMAND_MENU; 
	def.flags		= PF_ParamFlag_SUPERVISE | PF_ParamFlag_CANNOT_TIME_VARY;
	// can't use PF_PUI_STD_CONTROL_ONLY for the menu because AE won't handle
	// the undo stack automatically for us (?i think?)
//	def.ui_flags	= PF_PUI_STD_CONTROL_ONLY;
	def.param_type	= PF_Param_POPUP;
	def.u.pd.dephault = def.u.pd.value = kCmdRndChoose;
	def.u.pd.num_choices = kNumRandomizeMenuCommands;
	def.u.pd.u.namesptr = RANDOMIZE_CMD_MENU_ITEMS;		// just a pointer!  don't strcpy into this!
	PF_STRCPY(def.name, "Global Randomize");
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_RANDOM_VECTOR_LENGTH;
	// use PF_PUI_STD_CONTROL_ONLY because this control only has an effect when
	// you choose a menu item (i.e. it never has an immediate effect when you 
	// change it).   Thus we don't need the undo stack handling.  Note also that
	// we specify PF_ParamFlag_SUPERVISE.  AE will give a fatal error if we
	// don't set PF_ParamFlag_SUPERVISE because AE thinks (incorrectly) that a 
	// parameter with PF_PUI_STD_CONTROL_ONLY is useless unless it also has 
	// PF_ParamFlag_SUPERVISE set.  This is not the case, but we set 
	// PF_ParamFlag_SUPERVISE anyway just to keep AE happy.  We have no handler
	// for this parameter in the HandleChangedParam routine; we read the value out
	// of the params array when the user chooses the "explore" menu items.
	def.flags = PF_ParamFlag_START_COLLAPSED | PF_ParamFlag_CANNOT_TIME_VARY | PF_ParamFlag_SUPERVISE;
	def.ui_flags	= PF_PUI_STD_CONTROL_ONLY;
	PF_ADD_FLOAT_SLIDER("Explore Vector Length", 
						0.0f,				// VALID MIN
						10000000.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						1.0f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//=====================================
	// GLOBAL PARAMETERS
	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_GLOBAL_PARAMETERS_START; 
	def.param_type = PF_Param_GROUP_START;
	PF_STRCPY(def.name, "Global Parameters");
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_LOAD_PARAMETERS; 
	def.flags = PF_ParamFlag_SUPERVISE | PF_ParamFlag_CANNOT_TIME_VARY;		// send us a message when these param values are changed
	PF_ADD_SLIDER("Parameter Set", 0, 1000000, 0, 300, 0);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_CENTER_OFFSET_X; 
	PF_ADD_FLOAT_SLIDER("Shift X", 
						-1000000.0f,				// VALID MIN
						1000000.0f,			// VALID MAX
						-10.0f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						0.0f,				// DEFAULT
						9,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_CENTER_OFFSET_Y; 
	PF_ADD_FLOAT_SLIDER("Shift Y", 
						-1000000.0f,				// VALID MIN
						1000000.0f,			// VALID MAX
						-10.0f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						0.0f,				// DEFAULT
						9,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_ALPHA_BLEND; 
	def.u.fd.display_flags = 1;	// display as percent flag; not handled by macro.
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_FIXED("Alpha Blend", BLEND_MIN, BLEND_MAX, BLEND_MIN, BLEND_MAX, BLEND_DFLT, 1, 1);
			
	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_SHOW_ALPHA_ONLY;
	def.param_type = PF_Param_CHECKBOX;
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_STRCPY(def.name, "Show Alpha Only");
	def.u.bd.value = def.u.bd.dephault = FALSE;
	def.u.bd.u.nameptr = "Alpha Only";	/* this is strictly a pointer; don't STRCPY into it! */
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_BLEND; 
	def.u.fd.display_flags = 1;	// display as percent flag; not handled by macro.
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_FIXED("Transparency", BLEND_MIN, BLEND_MAX, BLEND_MIN, BLEND_MAX, BLEND_DFLT, 1, 1);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);		// AEF_PIXELS_PER_UNIT
	def.uu.id = DISKID_PIXELS_PER_UNIT; 
	def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
	PF_ADD_FLOAT_SLIDER("Pixels Per Unit", 
						0.0f,				// VALID MIN
						10000000.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						5000.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						100.0f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_ZOOM;
	// ABDTODO what are the proper bounds?
	// can this be negative?
	PF_ADD_FLOAT_SLIDER("Zoom", 
						-10000000.0f,				// VALID MIN
						10000000.0f,			// VALID MAX
						-10.0f,				// SLIDER MIN
						20.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						0.0f,				// DEFAULT
						9,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);
	def.uu.id = DISKID_SPATIAL_OVERSAMPLE;
	def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
	// ABDTODO what are the proper bounds?
	// can this be zero?
	PF_ADD_SLIDER("Spatial Oversample Factor", 1, 1000, 1, 20, 1);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_SPATIAL_FILTER_RADIUS;
	// ABDTODO what are the proper bounds?
	// can this be zero?
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_FLOAT_SLIDER("Spatial Filter Radius", 
						0.0f,				// VALID MIN
						10000.0f,			// VALID MAX
						0.00f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						0.5f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_SAMPLE_DENSITY;
	// ABDTODO what are the proper bounds?
	// can this be zero?
	PF_ADD_FLOAT_SLIDER("Sample Density", 
						0.0f,				// VALID MIN
						10000.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						200.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						1.0f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);		// AEF_GAMMA
	def.uu.id = DISKID_GAMMA;
	def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
	// ABDTODO what are the proper bounds?
	// can this be zero?
	PF_ADD_FLOAT_SLIDER("Gamma", 
						0.0f,				// VALID MIN
						10000.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						2.0f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_CONTRAST;
	// ABDTODO what are the proper bounds?
	// can this be zero?
	PF_ADD_FLOAT_SLIDER("Contrast", 
						0.0f,				// VALID MIN
						10000.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						1.0f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_BRIGHTNESS;
	PF_ADD_FLOAT_SLIDER("Brightness", 
						0.0f,				// VALID MIN
						10000.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						10.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						1.5f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_WHITE_LEVEL;
	// ABDTODO what are the proper bounds?
	// can this be zero or negative?
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_SLIDER("White Level", 0, 100000, 0, 1000, 200);

/*	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_RANDOM_PERIOCITY;
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_SLIDER("Random Periocity", 0, 1000000, 0, 100, 0);
*/
	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_COLORMAP;
	def.flags = PF_ParamFlag_SUPERVISE;
	def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
	// ABDTODO THIS should be popup but AE can't draw them w/big lists...

	// get the built-in cmaps
	gNumCmaps = NumCMaps()-1;

	// also include files in the cmap directory
	gNumCmaps += GetNthFile(CMAP_DIRECTORY, "*.MAP", -1, NULL);

	PF_ADD_SLIDER("Colormap #", 0, gNumCmaps, 0, gNumCmaps, 33);
/*	def.flags		= PF_ParamFlag_SUPERVISE;	// | PF_ParamFlag_CANNOT_TIME_VARY;
	def.param_type	= PF_Param_POPUP;
	def.u.sd.dephault = (short) def.u.sd.value = 33;
	def.u.sd.num_choices =	NumCMaps();
	def.u.sd.u.namesptr = COLOR_NAME_POPUP;	// "item1|item2|item3|item4"
	PF_STRCPY(def.name, "Colormap");
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;
*/
	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_COLORMAP_ROTATION;
	PF_ADD_FLOAT_SLIDER("Colormap Rotation", 
						0.0f,				// VALID MIN
						256.0f,			// VALID MAX
						0.0f,				// SLIDER MIN
						256.0f,				// SLIDER MAX	
						0.0f,		// CURVE TOLERANCE (not used for non-audio)
						0.0f,				// DEFAULT
						2,	// DISPLAY PRECISION
						0,		// DISPLAY FLAGS
						0);						// WANT PHASE

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_COLORMAP_INVERT;
	def.param_type = PF_Param_CHECKBOX;
	def.flags = 0;
	PF_STRCPY(def.name, "Colormap Invert");
	def.u.bd.value = def.u.bd.dephault = FALSE;
	def.u.bd.u.nameptr = "Inverted";	/* this is strictly a pointer; don't STRCPY into it! */
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_HUE_ROTATION;
	PF_ADD_ANGLE("Hue Rotation", 0);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_COLORMAP_INTERPOLATE;
	// WHAT ARE THE PROPER BOUNDS??
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_SLIDER("Colormap Interpolate", 0, 100, 0, 10, 0);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_NUM_BATCHES;
	def.flags = PF_ParamFlag_START_COLLAPSED;
	def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
	PF_ADD_SLIDER("Num Batches", 1, 10000, 1, 100, 1);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_NUM_STRIPS;
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_SLIDER("Num Strips", 0, 10000, 0, 100, 1);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);	
	def.uu.id = DISKID_RANDOM_SEED;
	def.flags = PF_ParamFlag_START_COLLAPSED;
	PF_ADD_SLIDER("Random Seed", 0, 32767, 0, 32767, 0);

	//-------------------------------------
	AEFX_CLR_STRUCT(def);
	def.uu.id = DISKID_GLOBAL_PARAMETERS_END;
	def.param_type = PF_Param_GROUP_END;
	if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

	//=====================================
	// THE XFORM GROUPS
	//-------------------------------------
	for ( xformNum = 0 ; xformNum < kNumXforms ; xformNum++ )
	{
		int i, colorDiskID, cmdDiskID, enabledDiskID, densityDiskID, groupStartDiskID, groupEndDiskID;
		int coeffDiskIDs[kNumCoefficients];
		int varDiskIDs[kNumVariations];
		//-------------------------------------
		AEFX_CLR_STRUCT(def);	
		switch (xformNum)
		{
			case 0 : 
				def.u.pd.u.namesptr = XFORM0_CMD_MENU_ITEMS; 	// just a pointer!  don't strcpy into this!
				colorDiskID = DISKID_XFORM0_COLOR;
				cmdDiskID = DISKID_XFORM0_COMMAND_MENU;
				enabledDiskID = -1;		// not used (always enabled, no user control)
				densityDiskID = DISKID_XFORM0_DENSITY;
				groupStartDiskID = DISKID_XFORM0_START;
				groupEndDiskID = DISKID_XFORM0_END;
				for ( i = 0 ; i < kNumVariations ; i++ )
					varDiskIDs[i] = DISKID_XFORM0_VAR0 + i;
				for ( i = 0 ; i < kNumCoefficients ; i++ )
					coeffDiskIDs[i] = DISKID_XFORM0_COEF0 + i;
				break;
			case 1 : 
				def.u.pd.u.namesptr = XFORM1_CMD_MENU_ITEMS; 	// just a pointer!  don't strcpy into this!
				colorDiskID = DISKID_XFORM1_COLOR;
				cmdDiskID = DISKID_XFORM1_COMMAND_MENU;
				enabledDiskID = -1;		// not used (always enabled, no user control)
				densityDiskID = DISKID_XFORM1_DENSITY;
				groupStartDiskID = DISKID_XFORM1_START;
				groupEndDiskID = DISKID_XFORM1_END;
				for ( i = 0 ; i < kNumVariations ; i++ )
					varDiskIDs[i] = DISKID_XFORM1_VAR0 + i;
				for ( i = 0 ; i < kNumCoefficients ; i++ )
					coeffDiskIDs[i] = DISKID_XFORM1_COEF0 + i;
				break;
			case 2 : 
				def.u.pd.u.namesptr = XFORM2_CMD_MENU_ITEMS; 	// just a pointer!  don't strcpy into this!
				colorDiskID = DISKID_XFORM2_COLOR;
				cmdDiskID = DISKID_XFORM2_COMMAND_MENU;
				enabledDiskID = DISKID_XFORM2_ENABLED;
				densityDiskID = DISKID_XFORM2_DENSITY;
				groupStartDiskID = DISKID_XFORM2_START;
				groupEndDiskID = DISKID_XFORM2_END;
				for ( i = 0 ; i < kNumVariations ; i++ )
					varDiskIDs[i] = DISKID_XFORM2_VAR0 + i;
				for ( i = 0 ; i < kNumCoefficients ; i++ )
					coeffDiskIDs[i] = DISKID_XFORM2_COEF0 + i;
				break;
			case 3 : 
				def.u.pd.u.namesptr = XFORM3_CMD_MENU_ITEMS; 	// just a pointer!  don't strcpy into this!
				colorDiskID = DISKID_XFORM3_COLOR;
				cmdDiskID = DISKID_XFORM3_COMMAND_MENU;
				enabledDiskID = DISKID_XFORM3_ENABLED;
				densityDiskID = DISKID_XFORM3_DENSITY;
				groupStartDiskID = DISKID_XFORM3_START;
				groupEndDiskID = DISKID_XFORM3_END;
				for ( i = 0 ; i < kNumVariations ; i++ )
					varDiskIDs[i] = DISKID_XFORM3_VAR0 + i;
				for ( i = 0 ; i < kNumCoefficients ; i++ )
					coeffDiskIDs[i] = DISKID_XFORM3_COEF0 + i;
				break;
			default: 
				// can't happen
				break;
		}

		
		def.uu.id = cmdDiskID;
		def.flags		= PF_ParamFlag_SUPERVISE | PF_ParamFlag_CANNOT_TIME_VARY;
	//	def.ui_flags	= PF_PUI_STD_CONTROL_ONLY;
		def.param_type	= PF_Param_POPUP;
		def.u.pd.dephault = def.u.pd.value = kCmdXformChoose;
		def.u.pd.num_choices = kNumXformMenuCommands;
		PF_SPRINTF(def.name, "Xform %ld Command", xformNum);
		if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

		//-------------------------------------
		// start this group
		//-------------------------------------
		AEFX_CLR_STRUCT(def);	
		def.uu.id = groupStartDiskID;
		def.param_type = PF_Param_GROUP_START;
		PF_SPRINTF(name, "Xform %ld", xformNum);
		PF_STRCPY(def.name, name);
		if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;

		//-------------------------------------
		if ((xformNum != 0) && (xformNum != 1))
		{
			AEFX_CLR_STRUCT(def);	
			def.uu.id = enabledDiskID;
			def.param_type = PF_Param_CHECKBOX;
			def.flags = 0;
			PF_SPRINTF(def.name, "Xform %ld Enabled", xformNum);
			def.u.bd.value = def.u.bd.dephault = FALSE;
			def.u.bd.u.nameptr = "Enabled";	/* this is strictly a pointer; don't STRCPY into it! */
			if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;
/*
			def.uu.id = enabledDiskID;
			PF_SPRINTF(name, "Xform %ld Enabled", xformNum);
			PF_ADD_SLIDER(name, 0, 1, 0, 1, 0);
*/		}

		//-------------------------------------
		AEFX_CLR_STRUCT(def);	
		def.uu.id = densityDiskID;
		PF_SPRINTF(name, "Xform %ld Density", xformNum);
		if ((xformNum == 0) || (xformNum == 1))
			PF_ADD_FLOAT_SLIDER(name, 
							0.0f,				// VALID MIN
							10000000.0f,		// VALID MAX
							0.0f,				// SLIDER MIN
							1.0f,				// SLIDER MAX	
							0.0f,		// CURVE TOLERANCE (not used for non-audio)
							1.0f,				// DEFAULT
							9,	// DISPLAY PRECISION
							0,		// DISPLAY FLAGS
							0);						// WANT PHASE
		else
			PF_ADD_FLOAT_SLIDER(name, 
							0.0f,				// VALID MIN
							10000000.0f,		// VALID MAX
							0.0f,				// SLIDER MIN
							1.0f,				// SLIDER MAX	
							0.0f,		// CURVE TOLERANCE (not used for non-audio)
							0.0f,				// DEFAULT
							9,	// DISPLAY PRECISION
							0,		// DISPLAY FLAGS
							0);						// WANT PHASE

		//-------------------------------------
		AEFX_CLR_STRUCT(def);	
		def.uu.id = colorDiskID;
		def.flags = PF_ParamFlag_START_COLLAPSED;
		PF_SPRINTF(name, "Xform %ld Color", xformNum);
		if (xformNum == 0)
			PF_ADD_FLOAT_SLIDER(name, 
							0.0f,				// VALID MIN
							1.0f,		// VALID MAX
							0.0f,				// SLIDER MIN
							1.0f,				// SLIDER MAX	
							0.0f,		// CURVE TOLERANCE (not used for non-audio)
							1.0f,				// DEFAULT
							9,	// DISPLAY PRECISION
							0,		// DISPLAY FLAGS
							0);						// WANT PHASE
		else
			PF_ADD_FLOAT_SLIDER(name, 
							0.0f,				// VALID MIN
							1.0f,		// VALID MAX
							0.0f,				// SLIDER MIN
							1.0f,				// SLIDER MAX	
							0.0f,		// CURVE TOLERANCE (not used for non-audio)
							0.0f,				// DEFAULT
							9,	// DISPLAY PRECISION
							0,		// DISPLAY FLAGS
							0);						// WANT PHASE

		// the seven variations
		for ( varNum = 0 ; varNum < kNumVariations ; varNum++ )
		{
			//-------------------------------------
			AEFX_CLR_STRUCT(def);	
			def.uu.id = varDiskIDs[varNum];
			if (varNum == 0)
				def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
			def.flags = PF_ParamFlag_SUPERVISE;		// send us a message when these param values are changed
			PF_SPRINTF(name, "Xform %ld Var %ld", xformNum, varNum);
	
			switch (varNum)
			{
				case 0 : strcat(name, " linear"); break;
				case 1 : strcat(name, " sinusoidal"); break;
				case 2 : strcat(name, " spherical"); break;
				case 3 : strcat(name, " swirl"); break;
				case 4 : strcat(name, " horseshoe"); break;
				case 5 : strcat(name, " polar"); break;
				case 6 : strcat(name, " bent"); break;
				default: break;
			}

			if (((varNum == 3) && (xformNum != 1)) || ((varNum == 4) && (xformNum == 1)))
				PF_ADD_FLOAT_SLIDER(name, 
									0.0f,				// VALID MIN
									100.0f,			// VALID MAX
									0.0f,				// SLIDER MIN
									100.0f,				// SLIDER MAX	
									0.0f,		// CURVE TOLERANCE (not used for non-audio)
									100.0f,				// DEFAULT
									2,	// DISPLAY PRECISION
									1,		// DISPLAY FLAGS
									0);						// WANT PHASE
			else
				PF_ADD_FLOAT_SLIDER(name, 
									0.0f,				// VALID MIN
									100.0f,			// VALID MAX
									0.0f,				// SLIDER MIN
									100.0f,				// SLIDER MAX	
									0.0f,		// CURVE TOLERANCE (not used for non-audio)
									0.0f,				// DEFAULT
									2,	// DISPLAY PRECISION
									1,		// DISPLAY FLAGS
									0);						// WANT PHASE
		}

		// the six coefficients
		for ( coeffNum = 0 ; coeffNum < kNumCoefficients ; coeffNum++ )
		{
			// coefficients from "rot3.place" xform 0
			double def0[] = {  0.466382, -0.06187, 0.0792416, 0.610638, -0.475657, -0.281155 };
			double def1[] = {  -0.513867, 0.271649, -0.254522, -0.550984, -0.674095, -0.600324 };

			//-------------------------------------
			AEFX_CLR_STRUCT(def);	
			def.uu.id = coeffDiskIDs[coeffNum];
			if (coeffNum == 0)
				def.ui_flags = PF_PUI_ECW_SEPARATOR;		// draw a separator line above this control
			PF_SPRINTF(name, "Xform %ld Coefficient %ld", xformNum, coeffNum);
			if (xformNum == 0)
			{
				PF_ADD_FLOAT_SLIDER(name, 
									-1.0f,				// VALID MIN
									1.0f,			// VALID MAX
									-1.0f,				// SLIDER MIN
									1.0f,				// SLIDER MAX	
									0.0f,		// CURVE TOLERANCE (not used for non-audio)
									(float)def0[coeffNum],				// DEFAULT
									9,	// DISPLAY PRECISION
									0,		// DISPLAY FLAGS
									0);						// WANT PHASE
			}
			else
			{
				PF_ADD_FLOAT_SLIDER(name, 
									-1.0f,				// VALID MIN
									1.0f,			// VALID MAX
									-1.0f,				// SLIDER MIN
									1.0f,				// SLIDER MAX	
									0.0f,		// CURVE TOLERANCE (not used for non-audio)
									(float)def1[coeffNum],				// DEFAULT
									9,	// DISPLAY PRECISION
									0,		// DISPLAY FLAGS
									0);						// WANT PHASE
			}
		}

		//-------------------------------------
		// end this group
		//-------------------------------------
		AEFX_CLR_STRUCT(def);
		def.uu.id = groupStartDiskID;
		def.param_type = PF_Param_GROUP_END;
		if (err = PF_ADD_PARAM(in_data, -1, &def)) return err;
	}

	//=====================================
	// NUM PARAMS
	//-------------------------------------
	// tell AE the number of params we have
	out_data->num_params = AEF_NUM_PARAMS;

	return err;
}

//=========================================================================
// returns number of lines in the file.
// pass full pathname to file in iFullFileName.
// assumes every line in the file is < 4000 characters
//-------------------------------------------------------------------------
int GetNumLines(const char *iFullFileName)
{
	FILE *f = fopen(iFullFileName, "r");
	if (f != NULL)
	{
		int numLines = 0;
		int done = FALSE;
		do
		{
			char buffer[4000];
			if (fgets(buffer, 4000, f) == NULL) 
				done = TRUE;
			else if ((buffer[0] != '\n') && (buffer[0] != '\0'))		// ignore blank lines
				numLines++;
		}
		while (!done);
		fclose(f);
		return numLines;
	}
	return 0;
}


//=========================================================================
// load cmap file n from the standard cmap dir
// typedef double clrmap[256][3];
//-------------------------------------------------------------------------
#if defined(macintosh)
#define MAX_PATH	255		// MAX_PATH is a Windows-specific constant, let's fake it for the Mac
#endif
void LoadColorMap(int n, clrmap c, int cmap_len)
{
	FILE *f = NULL;
	char fName[MAX_PATH];
	GetNthFile(CMAP_DIRECTORY, "*.MAP", n, fName);
	f = fopen(fName, "r");
	if (f != NULL)
	{
		int done = FALSE;
		int i;
		for( i = 0; i < 256; i++ )
		{
			int r, g, b;
			char buffer[256];
			if (fgets(buffer, 255, f) == NULL) 
				break;

			if (sscanf(buffer, "%d %d %d", &r, &g, &b) != 3)
			{
				 // error; must have 3 numbers on each line
				return;
			}
			c[i][0] = (double)r / 255.0;
			c[i][1] = (double)g / 255.0;
			c[i][2] = (double)b / 255.0;
		}
		fclose(f);
	}
}

//=========================================================================
// simple-minded error logging...
// call like this:
//	DBLOG("description of error", __FILE__, __LINE__);
//-------------------------------------------------------------------------
#define DBLOGFILE  "G:\\develop\\AEFlame\\dblog.txt"
//-------------------------------------------------------------------------
void DBLOG(const char *iString, const char *iFile, long iLine)
{
	if ((iString != NULL) && (iFile != NULL))
	{
		FILE *f;

		f = fopen(DBLOGFILE, "a+");
		if (f != NULL)
		{
			fprintf(f, "%s (%ld): %s\n", iFile, iLine, iString);
			fclose(f);
		}
	}
}

#if 0
//=========================================================================
// ABD this is a test routine, we create an image to test reading from.
// This is used to test our AE code without running any flame code.
// It is not used in the actual flame rendering at all; only for testing.
//-------------------------------------------------------------------------
unsigned char *TestImageAllocation(PF_InData *in_data, PF_OutData *out_data, int iWidth, int iHeight)
{
	long imageSize;
	unsigned char *image = NULL;

	// sanity check
	if ((iWidth <= 0) || (iHeight <= 0))
	{
		char str[256];
		PF_SPRINTF(str, "invalid parameters (width = %ld, height = %ld)", iWidth, iHeight);
		PF_STRCPY(out_data->return_msg, str);
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		DBLOG(str, __FILE__, __LINE__);
		goto all_done;
	}

	imageSize = kNumImageChannels * iWidth * iHeight;
	image = (unsigned char *) malloc(imageSize);

	if (image != NULL)
	{
		long col, row;
		unsigned char *imageDest = image;

		for ( row = 0 ; row < iHeight ; row++ ) 
		{
			for ( col = 0 ; col < iWidth ; col++ ) 
			{
				unsigned char r, g, b;

				r = (unsigned char)(((float)row / (float)iHeight) * 255.0f);
				g = (unsigned char)(((float)col / (float)iWidth) * 255.0f);
				b = (unsigned char)(((float)(row+col) / (float)(iHeight+iWidth)) * 255.0f);

				(*imageDest++) = r;
				(*imageDest++) = g;
				(*imageDest++) = b;
			}
		}
	}
	else
	{
		char str[256];
		PF_SPRINTF(str, "could not allocate image of %ld bytes (width = %ld, height = %ld) (not enough memory)", imageSize, iWidth, iHeight);
		PF_STRCPY(out_data->return_msg, str);
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		DBLOG(str, __FILE__, __LINE__);
	}

all_done:
	return image;
}
#endif


//=================================================================
// render a single frame using strips.
// ABDTODO eventually this routine should know about time....
// ABDTODO fix this, this routine doesn't work dammit
//-----------------------------------------------------------------
void RenderWithStrips(PF_InData *in_data,
					PF_OutData *out_data,
					PF_ParamDef *params[],
					unsigned char *iImage,
					frame_spec *iFrameSpec, 
					control_point *cps, 
					int iFirstFrame,
					int iLastFrame, 
					int iNumStrips,			// num strips; 0 means dynamically determined based on image height/width
					int iDeltaTime, 
					int iDoFields,
					int iTimeBase)
{
	if ((iImage != NULL)
		&& (iFirstFrame >= 0)
		&& (iLastFrame >= 0)
		&& (iNumStrips >= 0)			// 0 means dynamically determined
		&& (iFirstFrame <= iLastFrame)
		&& (iTimeBase >= 0))
	{
		int success;
		int time;
		int numStrips = iNumStrips;

		if (numStrips == 0)
		{
			numStrips = 10;		// ABDTODO make this compute something
		}

		for ( time = iFirstFrame ; time <= iLastFrame ; time += iDeltaTime ) 
		{
			int strip;
			double height = cps[0].height / (double) cps[0].pixels_per_unit;
			double center_base = cps[0].center[1] - ((numStrips-1)/2.0)*height/numStrips;

			for ( strip = 0 ; strip < numStrips ; strip++ ) 
			{
				unsigned char *strip_start;

				strip_start = iImage + (cps[0].height * strip * cps[0].width * 3);
				cps[0].center[1] = center_base + (height * (double) (strip / numStrips));

				iFrameSpec->time = (double) time;

				// note that we don't pass a progress routine to the render_rectangle
				// routine; instead, we update the AE progress ourselves after each strip.
				if (iDoFields) 
				{
					// even field
					success = render_rectangle(iFrameSpec, strip_start, cps[0].width, field_even, 3, 0);

					// increment time half a frame
					iFrameSpec->time += 0.5;

					// odd field
					success = render_rectangle(iFrameSpec, strip_start, cps[0].width, field_odd, 3, 0);
				} 
				else 
				{
					success = render_rectangle(iFrameSpec, strip_start, cps[0].width, field_both, 3, 0);
				}

				//---------
				// strip N is finished, show progres
				PF_PROGRESS(in_data, strip+1, numStrips);
			}
		}
	}
	else
	{
		// invalid parameters
	}
}

/*//=========================================================================
// This is the progress indicator callback for the flame rendering.  
// we are passed a double, 0.0 means we just started, 1.0 means we are done.
// This routine should display a progress meter.
// We call the AE progress routine.
// The progress goes from 0..1000 for the flame, and then from 1000..1000+height
// for the copying of the data into the AE buffer (and modification of the
// output pixels, if any)
//-------------------------------------------------------------------------
static PF_InData *gInData = NULL;
//-------------------------------------------------------------------------
void FlameProgressIndicator(double iProgress)
{
	if (gInData != NULL)
	{
		PF_InData *in_data = gInData;
		PF_Err err = PF_Err_NONE;

		if (err = PF_PROGRESS(in_data, (int)(iProgress * 1000.0), 1000)) 
		{
			// user tried to cancel.  too damn bad.
		}
	}
	else
	{
		// no way to draw progress bar because our global in_data pointer
		// is NULL...  this shouldn't happen
		DBLOG("gInData is NULL", __FILE__, __LINE__);
	}
}
*/
//=========================================================================
// 
//-------------------------------------------------------------------------
//=========================================================================
// 
//-------------------------------------------------------------------------
//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendGlobalParamsString(PF_InData	*in_data,
							  PF_ParamDef	*params[],
								char *s,
								int iPayAttentionToDownsample)
{
	char str[4000];
	double downsampleX = 1.0, downsampleY = 1.0, downsampleXY = 1.0;

	// we might not want to pay attention to downsample if we are creating
	// a string to save in a settings file instead of using for immediate
	// rendering.  
	if (iPayAttentionToDownsample)
	{
		// compute the downsample factors (used when computing reduced resolution images).
		// we need to use these value to modify the params that are pixel-based.
		downsampleX	= (double)in_data->downsample_x.num / (double)in_data->downsample_x.den;
		downsampleY	= (double)in_data->downsample_y.num / (double)in_data->downsample_y.den;

		// get a general downsample for pixel-based values that don't have separate XY components
		if (downsampleX > downsampleY)
			downsampleXY = downsampleX;
		else
			downsampleXY = downsampleY;
	}

	sprintf(str, "pixels_per_unit %.9g ", params[AEF_PIXELS_PER_UNIT]->u.fs_d.value * downsampleXY);
	strcat(s, str);
	sprintf(str, "center %.9g %.9g ", (double)params[AEF_CENTER_OFFSET_X]->u.fs_d.value, (double)params[AEF_CENTER_OFFSET_Y]->u.fs_d.value);
	strcat(s, str);
	sprintf(str, "gamma %.9g ", (double)params[AEF_GAMMA]->u.fs_d.value);
	strcat(s, str);
	sprintf(str, "spatial_filter_radius %.9g ", (double)params[AEF_SPATIAL_FILTER_RADIUS]->u.fs_d.value * downsampleXY);
	strcat(s, str);
	sprintf(str, "contrast %.9g ", (double)params[AEF_CONTRAST]->u.fs_d.value);
	strcat(s, str);
	sprintf(str, "brightness %.9g ", (double)params[AEF_BRIGHTNESS]->u.fs_d.value);
	strcat(s, str);
	sprintf(str, "zoom %.9g ", (double)params[AEF_ZOOM]->u.fs_d.value);		//  * downsampleXY
	strcat(s, str);
	sprintf(str, "spatial_oversample %ld ", params[AEF_SPATIAL_OVERSAMPLE]->u.sd.value);
	strcat(s, str);

/*	// check for LO-quality flag (for draft rendering) 
	if (in_data->quality == PF_Quality_LO)
	{
		// user has selected "LO" quality rendering....
		// reset various parameters that can make rendering take a long time...
		sprintf(str, "sample_density 3\n");
		strcat(s, str);
	}
	else
*/	{
		// normal quality... use the setting the user has selected
		sprintf(str, "sample_density %.9g ", (double)params[AEF_SAMPLE_DENSITY]->u.fs_d.value);
		strcat(s, str);
	}

	sprintf(str, "nbatches %ld ", params[AEF_NUM_BATCHES]->u.sd.value);
	strcat(s, str);
	sprintf(str, "white_level %ld ", params[AEF_WHITE_LEVEL]->u.sd.value);
	strcat(s, str);
	sprintf(str, "cmap_inter %ld ", params[AEF_COLORMAP_INTERPOLATE]->u.sd.value);
	strcat(s, str);
}
//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendXform0String(PF_ParamDef	*params[],
					  char *s)
{
	char str[4000];
	sprintf(str, "xform 0 density %.9g color %.9g ", params[AEF_XFORM0_DENSITY]->u.fs_d.value, params[AEF_XFORM0_COLOR]->u.fs_d.value);
	strcat(s, str);

	sprintf(str, "var %.9g %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)(params[AEF_XFORM0_VAR0]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM0_VAR1]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM0_VAR2]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM0_VAR3]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM0_VAR4]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM0_VAR5]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM0_VAR6]->u.fs_d.value) / 100.0f);
	//	strcpy(str, "var 0 0 0 1 0 0 0\n");
	strcat(s, str);

	sprintf(str, "coefs %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)params[AEF_XFORM0_COEF0]->u.fs_d.value, 
		(double)params[AEF_XFORM0_COEF1]->u.fs_d.value, 
		(double)params[AEF_XFORM0_COEF2]->u.fs_d.value, 
		(double)params[AEF_XFORM0_COEF3]->u.fs_d.value, 
		(double)params[AEF_XFORM0_COEF4]->u.fs_d.value, 
		(double)params[AEF_XFORM0_COEF5]->u.fs_d.value);
	strcat(s, str);
}
//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendXform1String(PF_ParamDef	*params[],
					  char *s)
{
	char str[4000];
	sprintf(str, "xform 1 density %.9g color %.9g ", params[AEF_XFORM1_DENSITY]->u.fs_d.value, params[AEF_XFORM1_COLOR]->u.fs_d.value);
	strcat(s, str);

	sprintf(str, "var %.9g %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)(params[AEF_XFORM1_VAR0]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM1_VAR1]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM1_VAR2]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM1_VAR3]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM1_VAR4]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM1_VAR5]->u.fs_d.value) / 100.0f, 
		(double)(params[AEF_XFORM1_VAR6]->u.fs_d.value) / 100.0f);
//	strcpy(str, "var 0 0 0 0 1 0 0\n");
	strcat(s, str);

	sprintf(str, "coefs %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)params[AEF_XFORM1_COEF0]->u.fs_d.value, 
		(double)params[AEF_XFORM1_COEF1]->u.fs_d.value, 
		(double)params[AEF_XFORM1_COEF2]->u.fs_d.value, 
		(double)params[AEF_XFORM1_COEF3]->u.fs_d.value, 
		(double)params[AEF_XFORM1_COEF4]->u.fs_d.value, 
		(double)params[AEF_XFORM1_COEF5]->u.fs_d.value);
	strcat(s, str);
}

//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendXform2String(PF_ParamDef	*params[],
					  char *s)
{
	char str[4000];
	sprintf(str, "xform 2 density %.9g color %.9g ", params[AEF_XFORM2_DENSITY]->u.fs_d.value, params[AEF_XFORM2_COLOR]->u.fs_d.value);
	strcat(s, str);

	sprintf(str, "var %.9g %.9g %.9g %.9g %.9g %.9g %.9g ", 
	(double)(params[AEF_XFORM2_VAR0]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM2_VAR1]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM2_VAR2]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM2_VAR3]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM2_VAR4]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM2_VAR5]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM2_VAR6]->u.fs_d.value) / 100.0f);
//		strcpy(str, "var 0 0 0 0 1 0 0\n");
	strcat(s, str);

	sprintf(str, "coefs %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)params[AEF_XFORM2_COEF0]->u.fs_d.value, 
		(double)params[AEF_XFORM2_COEF1]->u.fs_d.value, 
		(double)params[AEF_XFORM2_COEF2]->u.fs_d.value, 
		(double)params[AEF_XFORM2_COEF3]->u.fs_d.value, 
		(double)params[AEF_XFORM2_COEF4]->u.fs_d.value, 
		(double)params[AEF_XFORM2_COEF5]->u.fs_d.value);
	strcat(s, str);
}

//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendXform3String(PF_ParamDef	*params[],
					  char *s)
{
	char str[4000];
	sprintf(str, "xform 3 density %.9g color %.9g ", params[AEF_XFORM3_DENSITY]->u.fs_d.value, params[AEF_XFORM3_COLOR]->u.fs_d.value);
	strcat(s, str);

	sprintf(str, "var %.9g %.9g %.9g %.9g %.9g %.9g %.9g ", 
	(double)(params[AEF_XFORM3_VAR0]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM3_VAR1]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM3_VAR2]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM3_VAR3]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM3_VAR4]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM3_VAR5]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM3_VAR6]->u.fs_d.value) / 100.0f);
//		strcpy(str, "var 0 0 0 0 1 0 0\n");
	strcat(s, str);

	sprintf(str, "coefs %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)params[AEF_XFORM3_COEF0]->u.fs_d.value, 
		(double)params[AEF_XFORM3_COEF1]->u.fs_d.value, 
		(double)params[AEF_XFORM3_COEF2]->u.fs_d.value, 
		(double)params[AEF_XFORM3_COEF3]->u.fs_d.value, 
		(double)params[AEF_XFORM3_COEF4]->u.fs_d.value, 
		(double)params[AEF_XFORM3_COEF5]->u.fs_d.value);
	strcat(s, str);
}

/*
//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendXform4String(PF_ParamDef	*params[],
					  char *s)
{
	char str[4000];
	sprintf(str, "xform 4 density %.9g color %.9g ", params[AEF_XFORM4_DENSITY]->u.fs_d.value, params[AEF_XFORM4_COLOR]->u.fs_d.value);
	strcat(s, str);

	sprintf(str, "var %.9g %.9g %.9g %.9g %.9g %.9g %.9g ", 
	(double)(params[AEF_XFORM4_VAR0]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM4_VAR1]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM4_VAR2]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM4_VAR3]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM4_VAR4]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM4_VAR5]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM4_VAR6]->u.fs_d.value) / 100.0f);
//	strcpy(str, "var 0 0 0 0 1 0 0\n");
	strcat(s, str);

	sprintf(str, "coefs %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)params[AEF_XFORM4_COEF0]->u.fs_d.value, 
		(double)params[AEF_XFORM4_COEF1]->u.fs_d.value, 
		(double)params[AEF_XFORM4_COEF2]->u.fs_d.value, 
		(double)params[AEF_XFORM4_COEF3]->u.fs_d.value, 
		(double)params[AEF_XFORM4_COEF4]->u.fs_d.value, 
		(double)params[AEF_XFORM4_COEF5]->u.fs_d.value);
	strcat(s, str);
}


//=========================================================================
// 
//-------------------------------------------------------------------------
void AppendXform5String(PF_ParamDef	*params[],
					  char *s)
{
	char str[4000];
	sprintf(str, "xform 5 density %.9g color %.9g ", params[AEF_XFORM5_DENSITY]->u.fs_d.value, params[AEF_XFORM5_COLOR]->u.fs_d.value);
	strcat(s, str);

	sprintf(str, "var %.9g %.9g %.9g %.9g %.9g %.9g %.9g ", 
	(double)(params[AEF_XFORM5_VAR0]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM5_VAR1]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM5_VAR2]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM5_VAR3]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM5_VAR4]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM5_VAR5]->u.fs_d.value) / 100.0f, 
	(double)(params[AEF_XFORM5_VAR6]->u.fs_d.value) / 100.0f);
//	strcpy(str, "var 0 0 0 0 1 0 0\n");
	strcat(s, str);

	sprintf(str, "coefs %.9g %.9g %.9g %.9g %.9g %.9g ", 
		(double)params[AEF_XFORM5_COEF0]->u.fs_d.value, 
		(double)params[AEF_XFORM5_COEF1]->u.fs_d.value, 
		(double)params[AEF_XFORM5_COEF2]->u.fs_d.value, 
		(double)params[AEF_XFORM5_COEF3]->u.fs_d.value, 
		(double)params[AEF_XFORM5_COEF4]->u.fs_d.value, 
		(double)params[AEF_XFORM5_COEF5]->u.fs_d.value);
	strcat(s, str);
}
*/

//=========================================================================
// ABD this routine returns a buffer filled with a rendered image.  
// The buffer is a kNumImageChannels * iWidth * iHeight bytes large.
// the first 3 bytes are the R/G/B of the top-left corner of the image, etc.
// If there are any errors, NULL is returned.
// YOU MUST CALL free() WITH THE POINTER THAT IS RETURNED WHEN YOU ARE FINISHED
// WITH IT!  Or you will have a humongous memory leak...
//-------------------------------------------------------------------------
unsigned char *GetFrame(PF_InData	*in_data, 
						PF_OutData	*out_data, 
						PF_ParamDef	*params[],
						long iWidth, 
						long iHeight)
{
	int success;
	unsigned char *image = NULL;
	control_point templ;
	control_point *cps = NULL;
	int ncps = 0;
	frame_spec f;
	char s[4000] = "";
	char str[4000] = "";
	char *ss = NULL;

	// sanity check
	if (in_data == NULL)
	{
		DBLOG("in_data is NULL", __FILE__, __LINE__);
		PF_STRCPY(out_data->return_msg, "in_data is NULL");
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		goto all_done;
	}

	// sanity check
	if ((iWidth <= 0) || (iHeight <= 0))
	{
		char str[256];
		PF_SPRINTF(str, "invalid parameters (width = %ld, height = %ld)", iWidth, iHeight);
		DBLOG(str, __FILE__, __LINE__);
		PF_STRCPY(out_data->return_msg, str);
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		goto all_done;
	}

	cps = (control_point *) malloc(sizeof(control_point) * 300);
	if (cps == NULL)
	{
		// can't allocate cps... 
		// Apparently they ported After Effects to the Timex Sinclair.
		DBLOG("could not allocate cps buffer (not enough memory)", __FILE__, __LINE__);
		PF_STRCPY(out_data->return_msg, "could not allocate cps buffer (not enough memory)");
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		image = NULL;
		goto all_done;
	}

	/* init the template */
	// these are already downsampled for us automatically
	templ.width = iWidth;
	templ.height = iHeight;

	s[0] = '\0';

	// this end up getting ignored/reset, but what the hell
	strcpy(str, "image_size 450 450 ");
	strcat(s, str);

	AppendGlobalParamsString(in_data, params, s, 1);
	strcat(s, "; ");


/*	// original file
	image_size 450 450
	pixels_per_unit 150
	spatial_oversample 3
	spatial_filter_radius 1.0
	center -0.3 0.1
	gamma 3  brightness 2.3
	sample_density 8
	;
*/	
//	strcpy(s, "image_size 450 450\npixels_per_unit 150\nspatial_oversample 3\nspatial_filter_radius 1.0\ncenter -0.3 0.1\ngamma 3  brightness 2.3\nsample_density 8\n;\n");

	ss = s;
	parse_control_point(params, &ss, &templ);

	// Overwrite the template file's image height/width values... we 
	// absolutely *must* use the height/width specified by this
	// routine's parameters, not those that might specified in the 
	// template file... (because the calling routine assumes that the
	// buffer we are returning is going to have this width/height)
	templ.width = iWidth;
	templ.height = iHeight;


	//==================================================================
	// STARTING A NEW STRING...
	//==================================================================

	strcpy(s, "time 0 ");
	sprintf(str, "cmap %ld ", params[AEF_COLORMAP]->u.sd.value);
	strcat(s, str);
	//==================================================================
	AppendXform0String(params, s);
	//==================================================================
	AppendXform1String(params, s);
	//==================================================================
#if (kNumXforms > 2)
	if (params[AEF_XFORM2_ENABLED]->u.bd.value)
	{
		AppendXform2String(params, s);
	}
#endif
	//==================================================================
#if (kNumXforms > 3)
	if (params[AEF_XFORM3_ENABLED]->u.bd.value)
	{
		AppendXform3String(params, s);
	}
#endif
	//==================================================================
#if (kNumXforms > 4)
	if (params[AEF_XFORM4_ENABLED]->u.bd.value)
	{
		AppendXform4String(params, s);
	}
#endif
	//==================================================================
#if (kNumXforms > 5)
	if (params[AEF_XFORM5_ENABLED]->u.bd.value)
	{
		AppendXform5String(params, s);
	}
#endif
	//==================================================================
	strcat(s, ";\n");
	//==================================================================

/*	// original file
	time 0
	cmap 33
	xform 0 density 1 color 0
	var 0 0 0 1 0 0 0
	coefs 0.466382 -0.06187 0.0792416 0.610638 -0.475657 -0.281155
	xform 1 density 1 color 1
	var 0 0 0 0 1 0 0
	coefs -0.513867 0.271649 -0.254522 -0.550984 -0.674095 -0.600324
	;
*/
	
/*	// test case
	strcpy(s, "time 0\ncmap 33\nxform 0 density 1 color 0\nvar 0 0 0 1 0 0 0\ncoefs 0.466382 -0.06187 0.0792416 0.610638 -0.475657 -0.281155\nxform 1 density 1 color 1\nvar 0 0 0 0 1 0 0\ncoefs -0.513867 0.271649 -0.254522 -0.550984 -0.674095 -0.600324\n;\n");
*/	ss = s;
	cps[ncps] = templ;
	get_cmap(params, cmap_random, cps[ncps].cmap, CMAP_SIZE);
	parse_control_point(params, &ss, cps + ncps);
	ncps++;

/*	// test cases
	strcpy(s, "time 100\ncmap 33\nxform 0 density 1 color 0\nvar 0 0 0 1 0 0 0\ncoefs 0.06187 0.466382 -0.610638 0.0792416 -0.475657 -0.281155\nxform 1 density 1 color 1 var 0 0 0 0 1 0 0\ncoefs -0.271649 -0.513867 0.550984 -0.254522 -0.674095 -0.600324\n;\n");
	ss = s;
	cps[ncps] = templ;
	get_cmap(cmap_random, cps[ncps].cmap, CMAP_SIZE);
	parse_control_point(&ss, cps + ncps);
	ncps++;

	strcpy(s, "time 200\ncmap 33\nxform 0 density 1 color 0\nvar 0 0 0 1 0 0 0\ncoefs -0.466382 0.06187 -0.0792416 -0.610638 -0.475657 -0.281155\nxform 1 density 1 color 1\nvar 0 0 0 0 1 0 0\ncoefs 0.513867 -0.271649 0.254522 0.550984 -0.674095 -0.600324\n;\n");
	ss = s;
	cps[ncps] = templ;
	get_cmap(cmap_random, cps[ncps].cmap, CMAP_SIZE);
	parse_control_point(&ss, cps + ncps);
	ncps++;

	strcpy(s, "time 300\ncmap 33\nxform 0 density 1 color 0\nvar 0 0 0 1 0 0 0\ncoefs -0.06187 -0.466382 0.610638 -0.0792416 -0.475657 -0.281155\nxform 1 density 1 color 1\nvar 0 0 0 0 1 0 0\ncoefs 0.271649 0.513867 -0.550984 0.254522 -0.674095 -0.600324\n;\n");
	ss = s;
	cps[ncps] = templ;
	get_cmap(cmap_random, cps[ncps].cmap, CMAP_SIZE);
	parse_control_point(&ss, cps + ncps);
	ncps++;

	strcpy(s, "time 400\ncmap 33\nxform 0 density 1 color 0\nvar 0 0 0 1 0 0 0\ncoefs 0.466382 -0.06187 0.0792416 0.610638 -0.475657 -0.281155\nxform 1 density 1 color 1\nvar 0 0 0 0 1 0 0\ncoefs -0.513867 0.271649 -0.254522 -0.550984 -0.674095 -0.600324\n;\n");
	ss = s;
	cps[ncps] = templ;
	get_cmap(cmap_random, cps[ncps].cmap, CMAP_SIZE);
	parse_control_point(&ss, cps + ncps);
	ncps++;
*/

	// ABDTODO do we need to tweak this based on the frame rate???
	f.temporal_filter_radius = 0.5;

	
	f.cps = cps;
	f.ncps = ncps;
	f.time = (double) 0.0;

	// sanity checking....
	if (cps[0].width != iWidth)
	{
		DBLOG("cps[0].width != iWidth", __FILE__, __LINE__);
		PF_STRCPY(out_data->return_msg, "cps[0].width != iWidth");
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		image = NULL;
		goto all_done;
	}
	if (cps[0].height != iHeight)		// this is cps[1] in the original code.  should it be cps[0]?  does it matter?
	{
		DBLOG("cps[0].height != iHeight", __FILE__, __LINE__);
		PF_STRCPY(out_data->return_msg, "cps[0].height != iHeight");
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		image = NULL;
		goto all_done;
	}

	// allocate the image buffer
	// (that we will return)
	image = (unsigned char *) malloc(kNumImageChannels * cps[0].width * cps[0].height);		// this has cps[1].height in the original code.  should it be cps[0]?  does it matter?

	// bail if we couldn't allocate it...
	if (image == NULL)
	{
		DBLOG("could not allocate image buffer (not enough memory)", __FILE__, __LINE__);
		PF_STRCPY(out_data->return_msg, "could not allocate image buffer (not enough memory)");
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		goto all_done;
	}
	else
	{
		// everything is cool, we got our buffer...

		// before we render it, set the random seed so that we use the same
		// random sequence for each frame.  Otherwise, (i.e. with different
		// random sequence every frame) the images are noticably and annoyingly "sparkly"
		// especially at low sample densities.
		// You can try out the sparkly effect (i.e. having a different random sequence each frame)
		// by animating the "random seed" parameter.
		// It is much prettier IMHO to use the same random sequence every frame, so that
		// the random dots seem to be stationary on the overall fractal pattern instead of
		// jumping around.  
		srandom(params[AEF_RANDOM_SEED]->u.sd.value);

		// RENDER IT
		// ABDTODO fix the strip stuff
		// 1 strip means call render_rectangle directly
		// 0 strips means call RenderWithStrips, which will dynamically figure out an
		// 		appropriate num strip s
		// other numbers (> 1) mean use that exact number of strips.
		// unfortunately none of this strip stuff works right now...
		// when strips work 0 will be the best choice because it will figure out
		// what size of strip to use based on the resolution of the image (so that
		// we never run out of memory by trying to render huge strips)
		if (1)	// (params[AEF_NUM_STRIPS]->u.sd.value == 1)
		{
			success = render_rectangle(&f, image, cps[0].width, field_both, kNumImageChannels, 0);
			if (!success)
			{
				DBLOG("could not allocate buffer (not enough memory)", __FILE__, __LINE__);
				PF_STRCPY(out_data->return_msg, "could not allocate buffer (not enough memory)");
				out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
				goto all_done;
			}
		}
		else 
		{
			// ABDTODO this doesn't work
			// 
			RenderWithStrips(in_data, out_data, params, image, &f, cps,
						0,		// first frame
						0,		// last frame
						params[AEF_NUM_STRIPS]->u.sd.value,		// num strips (0 means dynamically determined)
						0,		// time delta
						0,		// do field rendering yes/no
						0);		// time base
		}
	}

all_done:
	if (cps != NULL)
		free(cps);	// remember to do this before exiting!   no leaks!

	// the caller has the responsibility to call free() on this return pointer...
	return image;
}


//=========================================================================
// This routine is our last opportunity to tweak the output colors/alpha channel.
// We are passed the computed output RGB(A), and the input RGBA.  We can use
// all of that info (and whatever else we care to consider) to modify the
// output RGB(A) values before they are stored in the output image.  
// If we don't care (i.e. if we just want the normal output image) then we
// do nothing.  We just leave the RGBA values as they are.
// Let's say we want to affect only the green channel, we could do:
//		*r = rIN;		// don't change the input red channel (i.e. copy the input pixel unchanged)
//		// let the green channel be changed...  (i.e. it already contains the new, computed output pixel)
//		*b = bIN;		// don't change the input blue channel (i.e. copy the input pixel unchanged)
// This routine *is* responsible, at the very least, for blending the input
// image and the computed imaged as per the user's setting of the "transparency"
// slider.   
//-------------------------------------------------------------------------
void ModifyColor(PF_InData *in_data,			// in_data pointer for access to more AE data
					PF_OutData *out_data,		// out_data ptr
					PF_ParamDef	*params[],		// params
					unsigned char *r,			// computed red component of output pixel
					unsigned char *g, 			// computed green component of output pixel
					unsigned char *b, 			// computed blue component of output pixel
					unsigned char *alpha,		// computed alpha if 4 channel; alpha of input pixel if 3-channel (i.e. same as aIN)
					unsigned char rIN,			// red component of corresponding input image pixel
					unsigned char gIN,			// green component of corresponding input image pixel
					unsigned char bIN,			// blue component of corresponding input image pixel
					unsigned char aIN)			// alpha component of corresponding input image pixel
{
	// r, g, b contain the computed RGB color for this pixel (0..255)
	// if this is a 4-channel image, alpha contains the computed alpha channel value
	// rIN, gIN, bIN contain the original pixels for the layer (0..255)
	// so we blend them or do whatever else we need to.


	if (params[AEF_SHOW_ALPHA_ONLY]->u.bd.value)
	{
		// display the output alpha channel (4-channel only)
		*r = *alpha;
		*g = *alpha;
		*b = *alpha;
		*alpha = aIN;
	}
	else
	{
		PF_Fixed alpha_blend = params[AEF_ALPHA_BLEND]->u.fd.value;

		if (alpha_blend == 0)
		{
			PF_Fixed transparency = params[AEF_BLEND]->u.fd.value;

			// no alpha (i.e. no alpha blending/ignore alpha channel)
			// 100% transparency means use input pixels only.
			// 0% transparency means use output pixels only.
			// in bewteen means calculate.
			if (transparency == (1L << 16)) 
			{
				// 100% transparency, simply copy the input pixels (ignore output pixels completely).
				*r = rIN;
				*g = gIN;
				*b = bIN;
			} 
			else if (transparency == 0) 
			{
				// 0% transparency, do nothing (rgb already contain the correct output pixels)
			}
			else
			{
				// in between blend, need to calculate.
				double blend = 1.0 - FIX2FLOAT(transparency);
				*r = rIN + (unsigned char)(blend * ((double)*r - (double)rIN));
				*g = gIN + (unsigned char)(blend * ((double)*g - (double)gIN));
				*b = bIN + (unsigned char)(blend * ((double)*b - (double)bIN));
			}
		}
		else
		{
			// *alpha contains the computed alpha for this pixel...
			// *alpha == 0 means use 100% of the input pixel
			// *alpha == 255 means use none of the input

			// in between; need to calculate

	/*		{
				double rgb[3];
				double hsv[3];

				rgb[0] = UCHAR2DOUBLE(*r);
				rgb[1] = UCHAR2DOUBLE(*g);
				rgb[2] = UCHAR2DOUBLE(*b);

				// rgb 0 - 1, h 0 - 6, s 0 - 1, v 0 - 1 
				// convert to hsv
				rgb2hsv(rgb, hsv);

		//		hsv[1] = 1.0;		// s
				hsv[2] = FIX2FLOAT(alpha_blend);		// v

				// convert back to rgb
				hsv2rgb(hsv, rgb);

				*r = DOUBLE2UCHAR(rgb[0]);
				*g = DOUBLE2UCHAR(rgb[1]);
				*b = DOUBLE2UCHAR(rgb[2]);
			}
	*/

			{
				double ablend = UCHAR2DOUBLE(*alpha);
				*r = rIN + (unsigned char)(ablend * ((double)*r - (double)rIN));
				*g = gIN + (unsigned char)(ablend * ((double)*g - (double)gIN));
				*b = bIN + (unsigned char)(ablend * ((double)*b - (double)bIN));
			}

			// because we have blended the images with the alpha ourselves, we need 
			// to set the output alpha to the input layer alpha (if we don't then 
			// the only part of the input layer which will be visible will be the
			// part where the fractal is, and we want the whole layer visible but
			// with the fractal alpha-blended onto it)  
			// Comment this line out if you don't understand what's happening and
			// you will see why this is necessary.   
			*alpha = aIN;
		}
	}
}



//=========================================================================
// called when AE wants us to render a frame.
//-------------------------------------------------------------------------
PF_Err	Render(	PF_InData		*in_data,
				PF_OutData		*out_data,
				PF_ParamDef		*params[],
				PF_LayerDef		*output)
{
	PF_Err		err = PF_Err_NONE;
			
	if (!gDisplayedAboutBox)
	{
		err = About(in_data, out_data, params, output);
		gDisplayedAboutBox = TRUE;
	}
	
	// clear pixels outside the extent_hint to zero alpha. 
	if (in_data->extent_hint.left != output->extent_hint.left ||
		in_data->extent_hint.top != output->extent_hint.top ||
		in_data->extent_hint.right != output->extent_hint.right ||
		in_data->extent_hint.bottom != output->extent_hint.bottom) 
	{
		err = PF_FILL(NULL, &output->extent_hint, output);
	}
	
	// geez, i hope PF_FILL didn't give us an error
	if (err == PF_Err_NONE) 
	{
		long		col, row, width, height, row_gutter_s, row_gutter_d;
		PF_Pixel	*in = NULL, *out = NULL;
		unsigned char *image = NULL;
		unsigned char *imageSrc = NULL;

		// get the starting row/col and the width/height of the output image
		col = in_data->extent_hint.left;
		row = in_data->extent_hint.top;
		width = in_data->extent_hint.right - col;
		height = in_data->extent_hint.bottom - row;

		// allocate and render an image.  
		// ABDTODO eventually it would be nice if these routines didn't have to 
		// allocate memory; they could write directly into the output image
		// to save us a blit step.  bah.  the fractal is so slow it makes very
		// little difference. :)

#define NON_FLAME_TEST_CASE 0
#if NON_FLAME_TEST_CASE
		// TEST CASE: render a frame without using flame...
		// This routine can be used in place of the actual rendering 
		// routine in order to test the rest of our code without 
		// worrying about possible bugs/leaks/etc in flame.
		// (or bugs that are a result of us calling the rendering
		// engine with incorrect params etc.)
		// TestImageAllocation returns a "rendered" image that we
		// can read out of just like we read out of the real one.
		// it is a simple 2D color ramp.
		image = TestImageAllocation(in_data, out_data, width, height);
#else

		// NORMAL CASE: render a frame using flame...
		// GetFrame allocates image with malloc()...
		image = GetFrame(in_data, out_data, params, width, height);
		
#endif // NON_FLAME_TEST_CASE

		// image is malloc'ed so we need to check the return ptr for NULL...
		if (image == NULL)
		{
			// crap, we couldn't allocate the image.
			// we could conceivably try to reduce the strip size
			// if this happens, so that we don't need such huge
			// buffers all at once.  For this version just bail.
			// Note no need to free(image) because image is NULL
			DBLOG("image is NULL", __FILE__, __LINE__);
			err = PF_Err_OUT_OF_MEMORY;
			return err;
		}

		// get a pointer that we can use to read the data out of image
		imageSrc = image;
		
		// magical incantation to determine the in and out pointers
		in = (PF_Pixel *)((char *)(params[AEF_DEFAULT_INPUT_LAYER]->u.ld.data + col) +
						  (row * params[AEF_DEFAULT_INPUT_LAYER]->u.ld.rowbytes));
		out = (PF_Pixel *)((char *)(output->data + col) + (row * output->rowbytes));
			 
		// magical incantation to determine the gutters
		row_gutter_d = output->rowbytes - ((long)width * sizeof(PF_Pixel));
		row_gutter_s = params[AEF_DEFAULT_INPUT_LAYER]->u.ld.rowbytes - ((long)width * sizeof(PF_Pixel));
	
		// loop through the entire output image, filling in the pixels...
		for (row = 0; row < height; row++) 
		{
			for (col = 0; col < width; col++) 
			{
				unsigned char r, g, b, alpha;

#define SUPER_SIMPLE_TEST_CASE 0
#if SUPER_SIMPLE_TEST_CASE

				// SUPER_SIMPLE_TEST_CASE:
				// simple test to see if our plugin routines work... (instead of renderng
				// an image and getting its pixels, compute the pixel colors directly)
				r = (unsigned char)(((float)row / (float)height) * 255.0f)
				g = (unsigned char)(((float)col / (float)width) * 255.0f)
				b = (unsigned char)(((float)(row+col) / (float)(height+width)) * 255.0f)

				// get the alpha channel value from the input image (since our
				// image has no alpha channel)
				alpha = in->alpha;
#else
				// NORMAL CASE: read the RGB color out of the rendered flame image.
				// note that this assumes that each component of the rendered image is an
				// unsigned char, and that there are exactly 3 components per pixel, that
				// the components are in R, G, B order, and that there are no gutters in the 
				// rendered image that we need to skip over.  

				// read the first 3 channels (which are always there)
				r = (*imageSrc++);
				g = (*imageSrc++);
				b = (*imageSrc++);

#if (kNumImageChannels == 4)

				// 4-channel image (i.e. RGBA)...
				// read the alpha out of the source image
				alpha = (*imageSrc++);
				
#elif (kNumImageChannels == 3)

				// 3-channel image (i.e. RGB)...
				// we don't read the alpha out of the image 
				// because there is no alpha channel in a 3-channel image...
				// So get the alpha channel value from the input image (since the
				// flame image has no alpha channel)
				alpha = in->alpha;
				
#else // some other num channels other then 3 or 4
// compile-time sanity check:  kNumImageChannels must be either 3 or 4,
// or else the code that assumes 3 or 4 channels will have to be modified.
#error kNumImageChannels must be either 3 or 4
#endif // (kNumImageChannels == 4)

#endif // SUPER_SIMPLE_TEST_CASE


				// at this point, r g and b contain the RGB components
				// of the output pixel, and alpha contains the alpha of the input image pixel.

				// modify the rgb values here (i.e. blend them with the "in" pixels)
				// (in->red, in->green, in->blue) or we could do tricky things like only store certain
				// hues/intensities/whatever into the final AE image buffer.  
				ModifyColor(in_data, out_data, params, &r, &g, &b, &alpha, in->red, in->green, in->blue, in->alpha);

				// store the RGBA value in our output AE image buffer...
				out->red = r;
				out->green = g;
				out->blue = b;
				out->alpha = alpha;

				// we are done with this pixel; 
				// increment the in and out pointers to point to the next pixel...
				in++;
				out++;
			}
			
			// we are done with this row;
			// increment the in and out pointers to point to the start of the next row of pixels...
			// (i.e. skip the gutters)
			in	= (PF_Pixel *)((char *)in + row_gutter_s);
			out = (PF_Pixel *)((char *)out + row_gutter_d);
	
			// update the progress bar each row (and check for user cancel...)
			// this is disabled here because by the time this part is called, the rendering
			// is finished and so it is pointless to cancel...
/*			if (err = PF_PROGRESS(in_data, row + 1, height)) 
				break;		// user cancelled, break out of loop
*/		}

		// all done copying the data from our allocated image 
		// into the AE output buffer...
		// absolutely MUST free() the image now or we'll have a huge memory leak....
		free(image);	
	}
	else
	{
		char str[256];
		PF_SPRINTF(str, "PF_FILL returned error %ld", err);
		PF_STRCPY(out_data->return_msg, str);
		out_data->out_flags |= PF_OutFlag_DISPLAY_ERROR_MESSAGE;
		DBLOG(str, __FILE__, __LINE__);
	}

	return err;
}





